import logging
import time

from .profile import Profile
from .question import Questions


log = logging.getLogger(__name__)


class Copy(object):
    """Copy photos, essays and other attributes from one profile to another."""

    copy_methods = ['photos', 'essays', 'looking_for', 'details', 'questions']

    def __init__(self, source_profile_or_user, dest_user):
        """
        :param source_profile_or_user: A :class:`~okcupyd.user.User` or
                                       :class:`~okcupyd.profile.Profile` object
                                       from which to copy attributes.
                                       :meth:`~.Copy.questions` will not
                                       will not preserve the importance of
                                       copied questions if a
                                       :class:`~okcupyd.profile.Profile`
                                       instance is provided.
        :param dest_user: A :class:`~okcupyd.user.User` to which data will be
                          copied
        """
        if isinstance(source_profile_or_user, Profile):
            self.source_profile = source_profile_or_user
            self.source_user = None
        else:
            self.source_user = source_profile_or_user
            self.source_profile = self.source_user.profile
        self.dest_user = dest_user

    def questions(self):
        """Copy questions to the destination user. When this class was
        initialized with a :class:`~okcupyd.profile.Profile`, this will
        delete any existing questions answers on the destination account.
        """
        if self.source_user:
            return self._copy_questions_from_user()
        else:
            return self._copy_questions_from_profile()

    def _copy_questions_from_user(self):
        dest_questions = self.dest_user.questions
        for key, importance in Questions.importance_name_to_number.items():
            questions = getattr(self.source_user.questions, key)
            for question in questions:
                log.debug(
                    dest_questions.respond_from_user_question(
                        question,
                        importance
                    ).content
                )

    def _copy_questions_from_profile(self):
        # Answer all of the questions that the source user has answered.
        # So that we can see their answers
        for question in self.source_profile.questions:
            log.debug(u'Answering {0}: {1} to see {2}\'s answer'.format(
                question.id, question.text, self.source_profile.username
            ))
            if not question.answered:
                self.dest_user.questions.respond(question.id, [1], [1], 3)

        # Load all of their questions. We use a new question fetchable because
        # the one that lives on the profile has been accessed and has cached
        # data.
        source_questions = self.source_profile.question_fetchable()[:]
        log.debug(u'Copying {0} questions from {1} to {2}'.format(
            len(source_questions),
            self.source_profile.username,
            self.dest_user.username
        ))
        id_to_user_question = {question.id: question
                               for question in self.dest_user.profile.questions}
        self.dest_user.questions.clear()
        while True:
            log.debug(u'Sleeping to wait for questions to clear')
            time.sleep(5)
            try:
                question = self.dest_user.profile.question_fetchable()[0]
            except IndexError:
                break
            else:
                log.debug(u'Destination user still has question {0}'.format(
                    question.text
                ))

        for question in source_questions:
            try:
                user_question = id_to_user_question[question.id]
            except KeyError:
                log.debug(u'No user question found for {0}: {1}'.format(
                    question.id, question.text
                ))
            else:
                log.debug(u'Answering {0}: {1}'.format(
                    question.id, question.text
                ))
                self.dest_user.questions.respond_from_question(question,
                                                               user_question, 3)

    def photos(self):
        """Copy photos to the destination user."""
        # Reverse because pictures appear in inverse chronological order.
        for photo_info in self.dest_user.profile.photo_infos:
            self.dest_user.photo.delete(photo_info)
        return [self.dest_user.photo.upload_and_confirm(info)
                for info in reversed(self.source_profile.photo_infos)]

    def essays(self):
        """Copy essays from the source profile to the destination profile."""
        for essay_name in self.dest_user.profile.essays.essay_names:
            setattr(self.dest_user.profile.essays, essay_name,
                    getattr(self.source_profile.essays, essay_name))

    def looking_for(self):
        """Copy looking for attributes from the source profile to the
        destination profile.
        """
        looking_for = self.source_profile.looking_for
        return self.dest_user.profile.looking_for.update(
            gentation=looking_for.gentation,
            single=looking_for.single,
            near_me=looking_for.near_me,
            kinds=looking_for.kinds,
            ages=looking_for.ages
        )

    def details(self):
        """Copy details from the source profile to the destination profile."""
        return self.dest_user.profile.details.convert_and_update(
            self.source_profile.details.as_dict
        )

    def all(self):
        """Invoke all of :meth:`~.Copy.questions`, :meth:`~.Copy.details`,
        :meth:`~.Copy.essays`, :meth:`~.Copy.photos`, :meth:`~.Copy.looking_for`
        """
        for method_name in self.copy_methods:
            getattr(self, method_name)()



import logging
import random
import time

import requests

from . import profile
from . import settings
from . import util
from .errors import AuthenticationError


log = logging.getLogger(__name__)


class Session(object):
    """A `requests.Session` with convenience methods for interacting with
    okcupid.com
    """

    default_login_headers = {
        'user-agent': ('Mozilla/5.0 (Macintosh; Intel Mac OS X 10_9_4) '
                       'AppleWebKit/537.36 (KHTML, like Gecko) '
                       'Chrome/37.0.2062.94 '
                       'Safari/537.36')
    }

    @classmethod
    def login(
            cls, username=None, password=None, requests_session=None,
            rate_limit=None
    ):
        """Get a session that has authenticated with okcupid.com.
        If no username and password is supplied, the ones stored in
        :class:`okcupyd.settings` will be used.

        :param username: The username to log in with.
        :type username: str
        :param password: The password to log in with.
        :type password: str
        :param rate_limit: Average time in seconds to wait between requests to OKC.
        :type rate_limit: float
        """
        requests_session = requests_session or requests.Session()
        session = cls(requests_session, rate_limit)
        # settings.USERNAME and settings.PASSWORD should not be made
        # the defaults to their respective arguments because doing so
        # would prevent this function from picking up any changes made
        # to those values after import time.
        username = username or settings.USERNAME
        password = password or settings.PASSWORD
        session.do_login(username, password)
        return session

    def __init__(self, requests_session, rate_limit=None):
        self._requests_session = requests_session
        self.log_in_name = None
        if isinstance(rate_limit, RateLimiter):
            self.rate_limiter = rate_limit
        else:
            self.rate_limiter = RateLimiter(rate_limit)

    def __getattr__(self, name):
        return getattr(self._requests_session, name)

    def do_login(self, username, password):
        credentials = {
            'username': username,
            'password': password,
            'okc_api': 1
        }
        login_response = self.okc_post('login',
                                       data=credentials,
                                       headers=self.default_login_headers,
                                       secure=True)
        login_json = login_response.json()
        log_in_name = login_json['screenname']
        if log_in_name is None:
            raise AuthenticationError(u'Could not log in as {0}'.format(username))
        if log_in_name.lower() != username.lower():
            log.warning(u'Expected to log in as {0} but '
                        u'got {1}'.format(username, log_in_name))
        log.debug(login_response.content.decode('utf8'))


        self.access_token = login_json.get("oauth_accesstoken")
        self.log_in_name = log_in_name
        self.headers.update(self.default_login_headers)

    def build_path(self, path, secure=None):
        if secure is None:
            secure = ('secure_login' in self.cookies and
                      int(self.cookies['secure_login']) != 0)
        return u'{0}://{1}/{2}'.format('https' if secure else 'http',
                                       util.DOMAIN, path)

    def get_profile(self, username):
        """Get the profile associated with the supplied username
        :param username: The username of the profile to retrieve."""
        return profile.Profile(self, username)

    def get_current_user_profile(self):
        """Get the `okcupyd.profile.Profile` associated with the supplied
        username.

        :param username: The username of the profile to retrieve.
        """
        return self.get_profile(self.log_in_name)


class RateLimiter(object):
    def __init__(self, rate_limit, wait_std_dev=None):
        self.rate_limit = rate_limit
        if rate_limit is not None and wait_std_dev is None:
            wait_std_dev = float(rate_limit) / 5
        self.wait_std_dev = wait_std_dev
        self.last_request = None

    def wait(self):
        if self.rate_limit is None: return
        if self.last_request is not None:
            wait_time = random.gauss(self.rate_limit, self.wait_std_dev)
            elapsed = time.time() - self.last_request
            if elapsed < wait_time:
                time.sleep(wait_time - elapsed)
        self.last_request = time.time()


def build_okc_method(method_name):
    def okc_method(self, path, secure=None, **kwargs):
        base_method = getattr(self, method_name)
        self.rate_limiter.wait()
        response = base_method(self.build_path(path, secure), **kwargs)
        response.raise_for_status()
        return response
    return okc_method


for method_name in ('get', 'put', 'post', 'delete'):
    setattr(Session, 'okc_{0}'.format(method_name), build_okc_method(method_name))



"""
Most of the collection objects that are returned from function
invocations in the okcupyd library are instances of
:class:`~okcupyd.util.fetchable.Fetchable`. In most cases, it is fine
to treat these objects as though they are lists because they can be iterated
over, sliced and accessed by index, just like lists:

.. code:: python

   for question in user.profile.questions:
       print(question.answer.text)

   a_random_question = user.profile.questions[2]
   for question in questions[2:4]:
       print(question.answer_options[0])

However, in some cases, it is important to be aware of the subtle
differences between :class:`~okcupyd.util.fetchable.Fetchable` objects
and python lists.
:class:`~okcupyd.util.fetchable.Fetchable` construct the elements that
they "contain" lazily. In most of its uses in the okcupyd library,
this means that http requests can be made to populate
:class:`~okcupyd.util.fetchable.Fetchable` instances as its elments
are requested.

The :attr:`~okcupyd.profile.Profile.questions`
:class:`~okcupyd.util.fetchable.Fetchable` that is used in the example
above fetches the pages that are used to construct its contents in
batches of 10 questions. This means that the actual call to retrieve
data is made when iteration starts. If you enable the request logger
when you run this code snippet, you get output that illustrates this
fact:

.. code::

   2014-10-29 04:25:04 Livien-MacbookAir requests.packages.urllib3.connectionpool[82461] DEBUG "GET /profile/ShrewdDrew/questions?leanmode=1&low=11 HTTP/1.1" 200 None
    Yes
    Yes
    Kiss someone.
    Yes.
    Yes
    Sex.
    Both equally
    No, I wouldn't give it as a gift.
    Maybe, I want to know all the important stuff.
    Once or twice a week
    2014-10-29 04:25:04 Livien-MacbookAir requests.packages.urllib3.connectionpool[82461] DEBUG "GET /profile/ShrewdDrew/questions?leanmode=1&low=21 HTTP/1.1" 200 None
    No.
    No
    No
    Yes
    Rarely / never
    Always.
    Discovering your shared interests
    The sun
    Acceptable.
    No.

Some fetchables will continue fetching content for quite a long time.
The search fetchable, for example, will fetch content until okcupid runs
out of search results. As such, things like:

.. code:: python

    for profile in user.search():
        profile.message("hey!")

should be avoided, as they are likely to generate a massive number of requests
to okcupid.com.


Another subtlety of the :class:`~okcupyd.util.fetchable.Fetchable`
class is that its instances cache its contained results. This means that
the second iteration over :attr:`okcupyd.profile.Profile.questions` in the
example below does not result in any http requests:

.. code:: python

    for question in user.profile.questions:
        print(question.text)

    for question in user.profile.questions:
        print(question.answer)

It is important to understand that this means that the contents of a
:class:`~okcupyd.util.fetchable.Fetchable` are not guarenteed to be in
sync with okcupid.com the second time they are requested. Calling
:meth:`~okcupyd.util.fetchable.Fetchable.refresh` will cause the
:class:`~okcupyd.util.fetchable.Fetchable` to request new data from
okcupid.com when its contents are requested. The code snippet that
follows prints out all the questions that the logged in user has
answered roughly once per hour, including ones that are answered while
the program is running.


.. code:: python

    import time

    while True:
        for question in user.profile.questions:
            print(question.text)
        user.profile.questions.refresh()
        time.sleep(3600)

Without the call to  `user.profile.questions.refresh()`, this program
would never update the user.profile.questions instance, and thus what
would be printed to the screen with each iteration of the for loop.
"""
import itertools
from lxml import html


class Fetchable(object):
    """List-like container object that lazily loads its contained items."""

    @classmethod
    def fetch_marshall(cls, fetcher, processor):
        return cls(FetchMarshall(fetcher, processor))

    def __init__(self, fetcher, **kwargs):
        """
        :param fetcher: An object with a `fetch` generator method that
                        retrieves items for the fetchable.
        :param nice_repr: Append the repr of a list containing the items that
                          have been fetched to this point by the fetcher.
                          Defaults to True.
        :param kwargs: Arguments that should be passed to the fetcher when it's
                       fetch method is called. These are stored on the fetchable
                       so they can be passed to the fetcher whenever
                       :meth:`refresh` is called.
        """
        self._fetcher = fetcher
        self._kwargs = kwargs
        self.refresh()

    def refresh(self, nice_repr=True, **kwargs):
        """
        :param nice_repr: Append the repr of a list containing the items that
                          have been fetched to this point by the fetcher.
        :type nice_repr: bool
        :param kwargs: kwargs that should be passed to the fetcher when its
                       fetch method is called. These are merged with the values
                       provided to the constructor, with the ones provided here
                       taking precedence if there is a conflict.
        """
        for key, value in self._kwargs.items():
            kwargs.setdefault(key, value)
        # No real good reason to hold on to this. DONT TOUCH.
        self._original_iterable = self._fetcher.fetch(**kwargs)
        self.exhausted = False
        if nice_repr:
            self._accumulated = []
            self._original_iterable = self._make_nice_repr_iterator(
                self._original_iterable, self._accumulated
            )
        else:
            self._accumulated = None
        self._clonable, = itertools.tee(self._original_iterable, 1)
        return self

    @staticmethod
    def _make_nice_repr_iterator(original_iterable, accumulator):
        for item in original_iterable:
            accumulator.append(item)
            yield item

    __call__ = refresh

    def __iter__(self):
        # This is hard to think about, but you can't ever use an iterator
        # if you plan on cloning it. Furthermore, you can only clone it once
        # (directly). For this reason, we throw away the iterator once it has
        # been cloned.
        new_iterable, self._clonable = itertools.tee(self._clonable, 2)
        return new_iterable

    def __getitem__(self, item):
        if not isinstance(item, slice):
            iterator = iter(self)
            assert isinstance(item, int)
            if item < 0:
                return list(iterator)[item]
            try:
                for i in range(item):
                    next(iterator)
                return next(iterator)
            except StopIteration:
                self.exhausted = True
                raise IndexError("The Fetchable does not have a value at the "
                                 "index that was provided.")
        return self._handle_slice(item)

    def _handle_slice(self, item):
        iterator = iter(self)
        # No point in being lazy if they want it all.
        if ((item.start is None and item.stop is None) or
            # If we have any negative numbers we have to expand the whole
            # thing anyway. This is also the case if there is no bound
            # on the slice, hence the `not item.stop` trigger.
            ((item.start and item.start < 0) or
             (not item.stop or item.stop < 0))):
            self.exausted = True
            return list(iterator)[item]
        accumulator = []
        # No need to do this for stop since we are sure it is not None.
        start = item.start or 0
        for _ in range(0, start, item.step or 1):
            try:
                next(iterator)
            except StopIteration: # This is strange but its what list do.
                self.exhausted = True
                break
        for i in range(item.stop - start):
            try:
                value = next(iterator)
            except StopIteration:
                self.exhausted = True
                break
            else:
                if item.step == None or i % item.step == 0:
                    accumulator.append(value)
        return accumulator

    def __repr__(self):
        fetched_type = repr(self._fetcher)
        if self._accumulated == None:
            list_repr = ''
        else:
            try:
                self[0]
            except:
                pass
            else:
                fetched_type = type(self._accumulated[0]).__name__
            list_repr = repr(self._accumulated)
            if not self.exhausted:
                if len(self._accumulated) == 0:
                    list_repr = '[...]'
                else:
                    list_repr = '{0}, ...]'.format(list_repr[:-1])

        return '<{0}[{1}]{2}>'.format(type(self).__name__,
                                    fetched_type, list_repr)

    def __len__(self):
        return len(self[:])

    def __add__(self, other):
        return self[:] + other[:]

    def __eq__(self, other):
        return self[:] == other[:]

    def __nonzero__(self):
        try:
            self[0]
        except IndexError:
            return False
        else:
            return True


class FetchMarshall(object):

    def __init__(self, fetcher, processor, terminator=None, start_at=1):
        self._fetcher = fetcher
        self._start_at = start_at
        self._processor = processor
        self._terminator = terminator or self.simple_decider

    @staticmethod
    def simple_decider(pos, last, text_response):
        return pos > last

    def fetch(self, start_at=None):
        pos = start_at or self._start_at
        while True:
            last = pos
            text_response = self._fetcher.fetch(start_at=pos)
            if not text_response: break
            for item in self._processor.process(text_response):
                if item is StopIteration:
                    raise StopIteration()
                yield item
                pos += 1
            if not self._terminator(pos, last, text_response):
                break

    def __repr__(self):
        return '{0}({1}, {2})'.format(type(self).__name__,
                                      repr(self._fetcher),
                                      repr(self._processor))


class SimpleProcessor(object):
    """Applies object_factory to each element found with element_xpath

    Accepts session merely to be consistent with the FetchMarshall interface.
    """

    def __init__(self, session, object_factory, element_xpath):
        self._object_factory = object_factory
        self._element_xpath = element_xpath

    def process(self, text_response):
        if not text_response.strip():
            yield StopIteration
            raise StopIteration()
        for element in self._element_xpath.apply_(
            html.fromstring(text_response)
        ):
            yield self._object_factory(element)

    def __repr__(self):
        return '<{0}({1}, {2})>'.format(type(self).__name__,
                                        repr(self._object_factory),
                                        repr(self._element_xpath))


class PaginationProcessor(object):

    def __init__(self, object_factory, element_xpb,
                 current_page_xpb, total_page_xpb):
        self._object_factory = object_factory
        self._element_xpb = element_xpb
        self._current_page_xpb = current_page_xpb
        self._total_page_xpb = total_page_xpb

    def _current_page(self, tree):
        return int(self._current_page_xpb.one_(tree))

    def _page_count(self, tree):
        return int(self._total_page_xpb.one_(tree))

    def _are_pages_left(self, tree):
        return self._current_page(tree) < self._page_count(tree)

    def process(self, text_response):
        tree = html.fromstring(text_response)
        for element in self._element_xpb.apply_(tree):
            yield self._object_factory(element)
        if not self._are_pages_left(tree):
            # This is pretty gross: Part of the processor protocol
            # is that if StopIteration is yielded, the loop above
            # will be terminated. No easy way around this short
            # of making bigger objects or abstracting less.
            yield StopIteration


class GETFetcher(object):

    def __init__(self, session, path, query_param_builder=lambda: {}):
        self._session = session
        self._path = path
        self._query_param_builder = query_param_builder

    def fetch(self, *args, **kwargs):
        response = self._session.okc_get(
            self._path, params=self._query_param_builder(*args, **kwargs)
        )
        return response.content.strip()

    def __repr__(self):
        return '<{0}("{1}")>'.format(type(self).__name__, self._path)



import inspect


class curry(object):
    """Curry a function or method.

    Applying :class:`~okcupyd.util.currying.curry` to a function
    creates a callable with the same functionality that can be invoked
    with an incomplete argument list to create a partial application
    of the original function.

    .. code-block:: python

        @curry
        def greater_than(x, y):
           return x > y

        >>> less_than_40 = greater_than(40)
        >>> less_than_40(39)
        True
        >>> less_than_40(50)
        False

    :class:`~okcupyd.util.currying.curry` allows functions to be partially
    invoked an arbitary number of times:

    .. code-block:: python

        @curry
        def add_5_things(a, b, c, d, e):
            return a + b + c + d + e

        # All of the following invocations of add_5_things
        >>> add_5_things(1)(1)(1)(1)(1)
        5

        one_left = add_5_things(1, 1)(3)(4) # A one place function that will
        # add 1 + 1 + 3 + 4 = 9 to whatever is provided as its argument.

        >>>> one_left(5)
        14
        >>> one_left(6)
        15

    A particular compelling use case for :class:`~okcupyd.util.currying.curry`
    is the creation of decorators that take optional arguments:

    .. code-block:: python

        @curry
        def add_n(function, n=1):
            def wrapped(*args, **kwargs):
                return function(*args, **kwargs) + n
            return wrapped

        @add_n(n=12)
        def multiply_plus_twelve(x, y):
            return x * y

        @add_n
        def multiply_plus_one(x, y):
            return x * y

        >>> multiply_plus_one(1, 1)
        2
        >>> multiply_plus_twelve(1, 1)
        13

    Notice that we were able to apply `add_n` regardless of whether or not an
    optional argument had been supplied earlier.

    The version of curry that is available for import has been curried itself.
    That is, its constructor can be invoked partially:

    .. code-block:: python

        @curry(evaluation_checker=lambda *args, **kwargs: len(args) > 2)
        def args_taking_function(*args):
            return reduce(lambda x, y: x*y, args)

        >>> args_taking_function(1, 2)
        2
        >>> args_taking_function(2)(3)
        6
        >>> args_taking_function(2, 2, 2, 2)
        16
    """

    @staticmethod
    def arity_evaluation_checker(function):
        """Build an evaluation checker that will return True when it is
        guaranteed that all positional arguments have been accounted for.
        """
        is_class = inspect.isclass(function)
        if is_class:
            function = function.__init__
        function_info = inspect.getargspec(function)
        function_args = function_info.args
        if is_class:
            # This is to handle the fact that self will get passed in
            # automatically.
            function_args = function_args[1:]

        def evaluation_checker(*args, **kwargs):
            kwarg_keys = set(kwargs.keys())
            if function_info.keywords is None:
                acceptable_kwargs = function_args[len(args):]
                # Make sure that we didn't get an argument we can't handle.
                if not kwarg_keys.issubset(acceptable_kwargs):
                    TypeError("Unrecognized Arguments: {0}".format(
                        [key for key in kwarg_keys
                         if key not in acceptable_kwargs]
                    ))

            needed_args = function_args[len(args):]
            if function_info.defaults:
                needed_args = needed_args[:-len(function_info.defaults)]
            return not needed_args or kwarg_keys.issuperset(needed_args)
        return evaluation_checker

    @staticmethod
    def count_evaluation_checker(count):
        def function(*args, **kwargs):
            return len(args) >= count
        return function

    def __init__(self, function, evaluation_checker=None,
                 args=(), kwargs=None, cache_name=False):
        """
        :param function: The function to curry.
        :evaluation checker: A function that controls when the function will
                             be evaluated. The evaluation checker recieves the
                             arguments that will be passed to the function when
                             making this decision.
        """
        self.function = function
        self.evaluation_checker = (evaluation_checker or
                                   self.arity_evaluation_checker(function))
        if cache_name is True:
            cache_name = self.function.__name__
        self.cache_name = cache_name
        self.args = args
        self.kwargs = kwargs or {}
        self.__name__ = function.__name__
        self.__doc__ = function.__doc__

    def __call__(self, *args, **kwargs):
        new_args = self.args + args
        new_kwargs = self.kwargs.copy()
        new_kwargs.update(kwargs)
        if self.evaluation_checker(*new_args, **new_kwargs):
            return self.function(*new_args, **new_kwargs)
        else:
            return type(self)(self.function, self.evaluation_checker,
                              new_args, new_kwargs)

    def __get__(self, obj, obj_type):
        if obj is None:
            return self
        bound = type(self)(self.function, self.evaluation_checker,
                           args=self.args + (obj,), kwargs=self.kwargs)
        # This caches the new partial application of the function on the
        # instance. Its not clear that this is a good idea but I'm leaving
        # it for now.
        if self.cache_name:
            setattr(obj, self.cache_name, bound)
        return bound

    def __repr__(self):
        return '<{0}.{1} of {2}>'.format(__name__, type(self).__name__,
                                         repr(self.function))


curry = curry(curry)



from sqlalchemy import Column
from sqlalchemy import ForeignKey
from sqlalchemy import Integer
from sqlalchemy.orm import relationship
from sqlalchemy.ext.orderinglist import ordering_list

from okcupyd.db import OKCBase


class MessageThread(OKCBase):

    __tablename__ = "message_thread"

    initiator_id = Column(Integer, ForeignKey("user.id"))
    initiator = relationship("User", foreign_keys=[initiator_id])

    respondent_id = Column(Integer, ForeignKey("user.id"))
    respondent = relationship("User", foreign_keys=[respondent_id])

    messages = relationship("Message", order_by="Message.thread_index",
                            collection_class=ordering_list('thread_index'),
                            backref='message_thread')



import itertools
import logging

import simplejson

from . import helpers
from . import magicnumbers
from . import util
from . import filter
from .profile import Profile
from .session import Session
from .xpath import xpb


log = logging.getLogger(__name__)

#: A :class:`~okcupyd.filter.Filters` object that stores all of the
#: filters that are accepted by :func:`.SearchFetchable`.
search_filters = filter.Filters()
search_filters.register_filter_builder(
    filter.gentation_filter,
    descriptions="The gentation of returned search results.",
    acceptable_values=magicnumbers.gentation_to_number.keys(),
    types=str
)
search_filters.register_filter_builder(
    filter.location_filter,
    types=int,
    descriptions=("The maximum distance from the specified location of "
                  "returned search results.")
)
search_filters.register_filter_builder(
    filter.age_filter,
    decider=filter.Filters.any_not_none_decider,
    types=(int, int),
    descriptions=['The minimum age of returned search results.',
                  'The maximum age of returned search results.']
)
search_filters.register_filter_builder(
    magicnumbers.get_kids_filter,
    keys=('has_kids', 'wants_kids'),
    acceptable_values=(magicnumbers.maps.has_kids.pattern_to_value.keys(),
                       magicnumbers.maps.wants_kids.pattern_to_value.keys()),
    decider=filter.Filters.any_decider
)
search_filters.register_filter_builder(
    magicnumbers.get_question_filter,
    decider=filter.Filters.any_not_none_decider,
    types=(':class:`~okcupyd.question.UserQuestion`', list),
    descriptions=["A question whose answer should be used to match search "
                  "results, or a question id. If a question id, "
                  "`question_answers` must be supplied.",
                  "A list of acceptable question answer indices."]
)
search_filters.register_filter_builder(
    magicnumbers.get_language_query,
    keys=('language',),
    acceptable_values=list(magicnumbers.language_map.keys())
)
search_filters.register_filter_builder(
    magicnumbers.get_join_date_filter,
    types=int,
    acceptable_values=magicnumbers.join_date_string_to_int.keys()
)


@search_filters.register_filter_builder(
    decider=search_filters.any_decider,
    types=(int, int),
    descriptions=["The minimum attractiveness of returned search results.",
                  "The maximum attractiveness of returned search results."]
)
def attractiveness_filter(attractiveness_min, attractiveness_max):
    if attractiveness_min == None:
        attractiveness_min = 0
    if attractiveness_max == None:
        attractiveness_max = 10000
    return '25,{0},{1}'.format(attractiveness_min, attractiveness_max)


@search_filters.register_filter_builder(
    types=int,
    descriptions=("The minimum number of questions answered by returned search "
                  "results."),
)
def question_count_filter(question_count_min):
    return '33,{0}'.format(question_count_min)


search_filters.register_filter_builder(
    magicnumbers.get_height_filter,
    descriptions=["The minimum height of returned search results.",

                  "The maximum height of returned search results."],
    acceptable_values=[["A height int in inches",
                       "An imperial height string e.g. 5'4\"",
                       "A metric height string e.g. 1.54m"] for _ in range(2)],
    decider=search_filters.any_not_none_decider
)


@search_filters.register_filter_builder(
    acceptable_values=('day', 'today', 'week', 'month', 'year', 'decade'),
    types=str
)
def last_online_filter(last_online):
    return '5,{0}'.format(helpers.format_last_online(last_online))


@search_filters.register_filter_builder(
    types=str,
    acceptable_values=('not single', 'married', 'single', 'any'),
    descriptions="The relationship status of returned search results."
)
def status_filter(status):
    status_int = 2  # single, default
    if status.lower() in ('not single', 'married'):
        status_int = 12
    elif status.lower() == 'any':
        status_int = 0
    return '35,{0}'.format(status_int)


def build_option_filter(key):
    @search_filters.register_filter_builder(
        keys=(key,),
        acceptable_values=magicnumbers.maps[key].pattern_to_value.keys(),
        types=str
    )
    @util.makelist_decorator
    def option_filter(value):
        return magicnumbers.filters[key](value)
for key in ['smokes', 'drinks', 'drugs', 'education_level', 'job',
            'income', 'religion', 'monogamy', 'diet', 'sign',
            'ethnicities', 'cats', 'dogs', 'bodytype']:
    build_option_filter(key)


class MatchCardExtractor(object):

    def __init__(self, div):
        self._div = div

    _id_xpb = xpb.button.with_classes(
        "binary_rating_button"
    ).select_attribute_('data-tuid')

    @property
    def id(self):
        return int(self._id_xpb.one_(self._div))

    @property
    def username(self):
        return xpb.div.with_class('username').get_text_(self._div).strip()

    @property
    def age(self):
        return int(xpb.span.with_class('age').get_text_(self._div))

    @property
    def location(self):
        return helpers.replace_chars(
            xpb.span.with_class('location').get_text_(self._div)
        )

    _match_percentage_xpb = xpb.div.with_classes('percentage_wrapper', 'match').\
                            span.with_classes('percentage')

    @property
    def match_percentage(self):
        try:
            return int(self._match_percentage_xpb.get_text_(self._div).strip('%'))
        except:
            return 0

    _enemy_percentage_xpb = xpb.div.with_classes('percentage_wrapper', 'enemy').\
                            span.with_classes('percentage').text_

    @property
    def enemy_percentage(self):
        try:
            return int(self._enemy_percentage_xpb.one_(self._div).strip('%'))
        except:
            return None

    @property
    def contacted(self):
        return bool(xpb.div.with_class('fancydate').apply_(self._div))

    @property
    def as_dict(self):
        return {
            # TODO(@IvanMalison): add rating.
            'username': self.username,
            'age': self.age,
            'location': self.location,
            'id': self.id,
            'contacted': self.contacted,
            'match_percentage': self.match_percentage,
            'enemy_percentage': self.enemy_percentage
        }


_match_card_xpb = xpb.div.with_classes('match_card')
# The docstring below is extended automatically. Read it in its entirety at
# http://okcupyd.readthedocs.org/en/latest/ or by generating the documentation
# yourself.
def SearchFetchable(session=None, **kwargs):
    """Search okcupid.com with the given parameters. Parameters are registered
    to this function through :meth:`~okcupyd.filter.Filters.register_filter_builder`
    of :data:`~okcupyd.html_search.search_filters`.

    :returns: A :class:`~okcupyd.util.fetchable.Fetchable` of
              :class:`~okcupyd.profile.Profile` instances.

    :param session: A logged in session.
    :type session: :class:`~okcupyd.session.Session`
    :param location: A location string which will be used to filter results.
    :param gender: The gender of the user performing the search.
    :param keywords: A list or space delimeted string of words to search for.
    :param order_by: The criteria to use for ordering results. expected_values:
                     'match', 'online', 'special_blend'
    """
    session = session or Session.login()
    return util.Fetchable.fetch_marshall(
        SearchHTMLFetcher(session, **kwargs),
        util.SimpleProcessor(
            session,
            lambda match_card_div: Profile(
                session=session,
                **MatchCardExtractor(match_card_div).as_dict
            ),
            _match_card_xpb
        )
    )


class SearchHTMLFetcher(object):

    _username_xpb = xpb.div.with_class('username')

    def __init__(self, session=None, **options):
        self._session = session or Session.login()
        self._options = options
        self.location = self._options.pop('location', None)
        self.gender = self._options.pop('gender', 'm')
        self.keywords = self._options.pop('keywords', None)
        self.order_by = self._options.pop('order_by', 'match').upper()
        self.count = self._options.pop('count', 9)
        self.filters = search_filters.legacy_build(**self._options)

    def _query_params(self, low=None):
        search_parameters = {
            'timekey': 1,
            'matchOrderBy': self.order_by.upper(),
            'custom_search': '0',
            'fromWhoOnline': '0',
            'mygender': self.gender,
            'update_prefs': '1',
            'sort_type': '0',
            'sa': '1',
            'count': self.count,
            'locid': (str(helpers.get_locid(self._session, self.location))
                      if self.location else 0),
            'ajax_load': 1,
            'discard_prefs': 1,
            'match_card_class': 'just_appended'
        }
        if low:
            search_parameters['low'] = low
        if self.keywords: search_parameters['keywords'] = self.keywords
        search_parameters.update(self.filters)
        return search_parameters

    def fetch(self, start_at=None, count=None):
        search_parameters = self._query_params(low=start_at)
        log.info(simplejson.dumps({'search_parameters': search_parameters}))
        response = self._session.okc_get('match',
                                         params=search_parameters)
        try:
            html_search = response.json()['html']
        except:
            log.warning(simplejson.dumps({'failure': response.content}))
            raise
        return html_search

    def __unicode__(self):
        return u'{0}({1})'.format(type(self).__name__, repr(self._options))

    __repr__ = __unicode__


def search(session=None, count=1, **kwargs):
    return SearchFetchable(session, count=count, **kwargs)[:count]


search_filters.add_to_docstring_of(SearchFetchable)



import logging
import os

from sqlalchemy import Column, DateTime, Integer
from sqlalchemy import create_engine
from sqlalchemy.exc import IntegrityError
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import ColumnProperty, Query, class_mapper, sessionmaker
from sqlalchemy.sql import func
from wrapt import decorator

from . import types


log = logging.getLogger(__name__)


class Query(Query):
    pass


echo = False


engine = create_engine('sqlite://', convert_unicode=True, echo=True)
Session = sessionmaker(
    autoflush=False,
    expire_on_commit=False,
    bind=engine,
    query_cls=Query
)


class txn(object):

    def __init__(self, session_class=None):
        self.session_class = session_class or Session

    def __call__(self, func):
        def wrapped(*args, **kwargs):
            with self:
                return func(self.session, *args, **kwargs)
        return wrapped

    def __enter__(self):
        self.session = self.session_class()
        return self.session

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_val:
            self.session.rollback()

            if isinstance(exc_val, Exception):
                raise
            else:
                raise exc_type(exc_val)
        else:
            self.session.commit()
            self.session.close()


@decorator
def with_txn(function, instance, args, kwargs):
    with txn() as session:
        return function(session, *args, **kwargs)


class Base(object):

    id = Column(Integer, primary_key=True)
    created_at = Column(DateTime, nullable=False, default=func.now())

    def upsert_model(self, id_key='id'):
        return self.upsert([self], id_key=id_key)

    @classmethod
    def columns(cls):
        return [prop for prop in class_mapper(cls).iterate_properties
                if isinstance(prop, ColumnProperty)]

    @classmethod
    def upsert_no_txn(cls, session, models, id_key='id'):
        ids_to_look_for = [getattr(model, id_key) for model in models]
        found_models = cls.find_all_no_txn(session, ids_to_look_for,
                                           id_key=id_key)
        id_to_model = {getattr(model, id_key): model for model in found_models}
        try:
            key_type = type(next(iter(id_to_model.keys())))
        except:
            key_type = int
        for model in models:
            model_id = key_type(getattr(model, id_key))
            if model_id in id_to_model:
                # Get the primary key in a better way here
                model.id = id_to_model[model_id].id
                session.merge(model)
            else:
                id_to_model[model_id] = model
                session.add(model)
        log.info(id_to_model)
        return id_to_model

    @classmethod
    def upsert_one_no_txn(cls, session, model, **kwargs):
        return next(iter(cls.upsert_no_txn(session, [model], **kwargs).values()))

    upsert = with_txn(upsert_no_txn)

    @classmethod
    def safe_upsert(cls, *args, **kwargs):
        # TODO(imalison): use a retry decorator here.
        try:
            return cls.upsert(*args, **kwargs)
        except IntegrityError:
            return cls.upsert(*args, **kwargs)

    @classmethod
    def upsert_okc(cls, model, **kwargs):
        return next(iter(cls.safe_upsert([model], id_key='okc_id').values()))

    @classmethod
    def find_all_no_txn(cls, session, identifiers, id_key='id'):
        return cls.find_query(session, identifiers, id_key=id_key).all()

    find_all = with_txn(find_all_no_txn)

    @classmethod
    def find_no_txn(cls, session, identifier, id_key='id'):
        return cls.find_query(session, [identifier], id_key=id_key).one()

    find = with_txn(find_no_txn)

    @classmethod
    def query_no_txn(cls, session, *args, **kwargs):
        if not kwargs and len(args) == 1 and isinstance(args[0], (int, str)):
            args = cls.id == args[0],
        return cls.build_query(session, *args, **kwargs).all()

    query = with_txn(query_no_txn)

    @classmethod
    def find_query(cls, session, identifiers, id_key='id'):
        return session.query(cls).filter(getattr(cls, id_key).in_(identifiers))

    @classmethod
    def build_query(cls, session, *args, **kwargs):
        return session.query(cls).filter(*args).filter_by(**kwargs)


Base = declarative_base(engine, cls=Base)


class OKCBase(Base):

    __abstract__ = True

    okc_id = Column(types.StringBackedInteger, nullable=False, unique=True)


def reset_engine(engine):
    Session.configure(bind=engine)
    Base.metadata.bind = engine
    return engine


def set_sqlite_db_file(file_path):
    return reset_engine(create_engine('sqlite:///{0}'.format(file_path),
                                      convert_unicode=True,
                                      echo=echo))


database_uri = os.path.join(os.path.dirname(__file__), 'okcupyd.db')
set_sqlite_db_file(database_uri)



from sqlalchemy import Column, DateTime, ForeignKey, Integer, String
from sqlalchemy.orm import backref, relationship

from okcupyd.db import Base, OKCBase


class User(OKCBase):

    __tablename__ = 'user'

    @classmethod
    def from_profile(cls, profile):
        return cls(okc_id=profile.id, handle=profile.username, age=profile.age,
                   location=profile.location)

    handle = Column(String, nullable=False)
    age = Column(String, nullable=False)
    location = Column(String, nullable=False)


class OKCupydUser(Base):

    __tablename__ = 'okcupyd_user'

    inbox_last_updated = Column(DateTime, nullable=True)
    outbox_last_updated = Column(DateTime, nullable=True)
    user_id = Column(Integer, ForeignKey("user.id"), unique=True)
    user = relationship("User", foreign_keys=[user_id],
                        backref=backref('okcupyd_user', uselist=False))



import logging

from invoke import Collection
import IPython

from . import copy
from . import db as db_collection
from . import util as task_util
from okcupyd import user, db, util


log = logging.getLogger(__name__)

ns = Collection()

ns.add_collection(copy)
ns.add_collection(db_collection)
nstask = task_util.build_task_factory(ns)


@nstask(aliases='l')
def login(ctx):
    util.get_credentials()


@nstask(pre=(login,), default=True, aliases='i')
def interactive(ctx):
    u = user.User()
    u = u
    IPython.embed()


@nstask(aliases='s')
def session(ctx):
    with db.txn() as session:
        session = session
        IPython.embed()


@nstask(pre=(login,))
def enable_logger(logger_name):
    util.enable_logger(logger_name)


@nstask(aliases='c')
def credentials(ctx, module_name):
    util.update_settings_with_module(module_name)


@nstask(aliases=('eal',))
def enable_all_loggers(ctx):
    for logger_name in ('okcupyd', 'requests', __name__):
        util.enable_logger(logger_name)
    db.Session.kw['bind'].echo = True



import datetime
import logging
import itertools
import time

import mock
import contextlib2 as contextlib
import pytest
from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from vcr.cassette import Cassette, CassetteContextDecorator

from . import util
from okcupyd import db, settings
from okcupyd import util as okcupyd_util
from okcupyd.db import model, adapters


log = logging.getLogger(__name__)
BREAK_ON_EXCEPTION = False


def pytest_addoption(parser):
    okcupyd_util.add_command_line_options(
        parser.addoption, use_short_options=False
    )
    parser.addoption('--live', dest='skip_vcrpy', action='store_true',
                     default=False, help="Skip the patching of http requests in"
                     " tests. USE WITH CAUTION. This will interact with the "
                     "okcupid website and send messages with any provided user "
                     "credentials.")
    parser.addoption('--scrub', dest='scrub', action='store_true',
                     help="Scrub PII from http requests/responses. "
                     "This is useful for recording cassetes.", default=False)
    parser.addoption('--resave', dest='resave',
                     action='store_true', default=False,
                     help="Resave cassettes. Use to retoractively scrub "
                     "cassettes.")
    parser.addoption('--cassette-mode', dest='cassette_mode', action='store',
                     default='once', help="Accept new requests in all tests.")
    parser.addoption('--record', dest='record', action='store_true',
                     default=False, help="Re-record cassettes for all tests.")
    parser.addoption('--break-exc', dest='break_on_exception',
                     default=False, action='store_true', help="Enter ipdb if an"
                     " exception is encountered in a test.")


def patch(option, *patches, **kwargs):
    negate = kwargs.get('negate', False)

    @pytest.yield_fixture(autouse=True, scope='session')
    def patch_conditionally(request):
        condition = bool(request.config.getoption(option))
        if negate:
            condition = not condition
        if condition:
            with contextlib.ExitStack() as exit_stack:
                for patch in patches:
                    exit_stack.enter_context(patch)
                yield
        else:
            yield

    return patch_conditionally


patch_settings = patch('credential_files',
                       mock.patch.object(settings, 'USERNAME', 'username'),
                       mock.patch.object(settings, 'PASSWORD', 'password'),
                       mock.patch.object(settings, 'AF_USERNAME', 'username'),
                       mock.patch.object(settings, 'AF_PASSWORD', 'password'),
                       negate=True)


patch_save = patch(
    'resave', mock.patch.object(
        Cassette, '_save', okcupyd_util.curry(Cassette._save)(force=True)
    )
)


def reraise_exception(obj, error_type, error, traceback):
    if error is not None:
        raise


patch_use_cassette_enabled = patch(
    'skip_vcrpy',
    mock.patch.object(CassetteContextDecorator, '__enter__',
                      lambda stuff: log.debug("Skipping vcrpy patching.")),
    mock.patch.object(CassetteContextDecorator, '__exit__',
                      reraise_exception)
)

patch_vcrpy_filters = patch(
    'scrub', mock.patch.object(util, 'SHOULD_SCRUB', True)
)

patch_break_on_exception = patch(
    'break_on_exception',
    mock.patch('okcupyd_testing.conftest.BREAK_ON_EXCEPTION', True)
)

original_load = Cassette._load


def new_load(self):
    self.dirty = False
    self.rewound = True


patch_record = patch(
    'record', mock.patch.object(
        Cassette, '_load', new_load
    )
)


@pytest.fixture(autouse=True, scope='session')
def set_vcr_options(request):
    util.okcupyd_vcr.record_mode = request.config.getoption('cassette_mode')


@pytest.fixture(autouse=True, scope='session')
def process_command_line_args(request):
    okcupyd_util.handle_command_line_options(request.config.option)


@pytest.fixture(scope='function')
def engine(request):
    return create_engine('sqlite://', echo=request.config.getoption('echo'))


@pytest.fixture(scope='function')
def session(engine):
    new_kwargs = db.Session.kw.copy()
    new_kwargs['bind'] = engine
    return sessionmaker(**new_kwargs)


@pytest.yield_fixture(autouse=True, scope='function')
def setup_db(engine, session):
    with mock.patch.object(db, 'Session', session):
        orig_engine = db.Base.metadata.bind
        db.Base.metadata.bind = engine
        db.Base.metadata.drop_all()
        db.Base.metadata.create_all()
        yield
        db.Base.metadata.bind = orig_engine


@pytest.fixture
def T(mock_profile_builder, mock_message_thread_builder, mock_message_builder):

    class testing(object):
        pass

    testing.ensure = mock.Mock()
    testing.build_mock = mock.Mock()
    testing.factory = mock.Mock()

    def ensure_thread_model_resembles_okcupyd_thread(
        thread_model, okcupyd_thread
    ):
        assert thread_model.okc_id == okcupyd_thread.id
        ensure_user_model_resembles_okcupyd_profile(
            thread_model.initiator, okcupyd_thread.initiator
        )
        ensure_user_model_resembles_okcupyd_profile(
            thread_model.respondent, okcupyd_thread.respondent
        )
        for pair in zip(thread_model.messages, okcupyd_thread.messages):
            ensure_message_model_resembles_okcupyd_message(*pair)
            assert len(thread_model.messages) == len(okcupyd_thread.messages)

    def ensure_message_model_resembles_okcupyd_message(
        message_model, okcupyd_message
    ):
        assert message_model.okc_id == okcupyd_message.id
        assert message_model.sender.handle == okcupyd_message.sender.username
        assert message_model.recipient.handle == okcupyd_message.recipient.username
        assert message_model.text == okcupyd_message.content

    def ensure_user_model_resembles_okcupyd_profile(user_model,
                                                    okcupyd_profile):
        assert user_model.handle == okcupyd_profile.username

    testing.ensure.user_model_resembles_okcupyd_profile = \
        ensure_user_model_resembles_okcupyd_profile
    testing.ensure.message_model_resembles_okcupyd_message = \
        ensure_message_model_resembles_okcupyd_message
    testing.ensure.thread_model_resembles_okcupyd_thread = \
        ensure_thread_model_resembles_okcupyd_thread

    testing.build_mock.thread = mock_message_thread_builder
    testing.build_mock.message = mock_message_builder
    testing.build_mock.profile = mock_profile_builder

    def build_message_thread():
        message_thread = testing.build_mock.thread()
        return adapters.ThreadAdapter(message_thread).get_thread()[0]

    def build_user(username):
        profile = testing.build_mock.profile(username)
        return adapters.UserAdapter(profile).get()

    def build_okcupyd_user(user):
        user_model = model.User.from_profile(user.profile)
        user_model.upsert_model(id_key='okc_id')
        okcupyd_user = model.OKCupydUser(user_id=user_model.id)
        okcupyd_user.upsert_model(id_key='user_id')
        return okcupyd_user

    testing.factory.message_thread = build_message_thread
    testing.factory.user = build_user
    testing.factory.okcupyd_user = build_okcupyd_user
    return testing


@pytest.fixture
def mock_profile_builder():
    counter = itertools.count()
    next(counter)
    username_to_profile = {}

    def _build_mock_profile(username='username', age=30, id=None,
                            location='San Francisco, CA', **kwargs):
        if username in username_to_profile:
            return username_to_profile[username]
        if id is None:
            id = next(counter)
        mock_profile = mock.MagicMock(id=id or next(counter), location=location,
                                      age=age, username=username, **kwargs)
        username_to_profile[username] = mock_profile
        return mock_profile

    return _build_mock_profile


@pytest.fixture
def mock_message_builder():
    counter = itertools.count()
    next(counter)

    def _build_mock_message(id=None, sender='sender', recipient='recipient',
                            content='content', **kwargs):
        kwargs.setdefault('time_sent',
                          datetime.datetime(year=2014, day=2, month=4))
        if id is None:
            id = next(counter)
        assert isinstance(sender, str)
        assert isinstance(recipient, str)
        return mock.MagicMock(id=id, sender=mock.Mock(username=sender),
                              recipient=mock.Mock(username=recipient),
                              content=content, **kwargs)

    return _build_mock_message


@pytest.fixture
def mock_message_thread_builder(mock_message_builder, mock_profile_builder):
    counter = itertools.count()
    next(counter)

    def _build_mock_message_thread(id=None, message_count=2,
                                   initiator='initiator',
                                   respondent='respondent', **kwargs):
        if id is None:
            id = next(counter)
        messages = [mock_message_builder(content='{0}'.format(i),
                                         sender=respondent
                                         if i % 2 else initiator,
                                         recipient=initiator
                                         if i % 2 else respondent)
                    for i in range(message_count)]
        kwargs.setdefault('datetime',
                          datetime.datetime(year=2014, day=2, month=4))
        message_thread = mock.MagicMock(id=id, messages=messages, **kwargs)
        message_thread.initiator = mock_profile_builder(initiator)
        message_thread.respondent = mock_profile_builder(respondent)
        return message_thread
    return _build_mock_message_thread


def pytest_exception_interact():
    if BREAK_ON_EXCEPTION:
        import ipdb; ipdb.set_trace()
    try:
        models = []
        for key in dir(model):
            prospect = getattr(model, key)
            if isinstance(prospect, type) and issubclass(prospect, db.Base):
                models.append(prospect)
        model_class_to_instances = {}
        for model_class in models:
            model_class_to_instances[model_class] = {
                instance.okc_id: instance for instance in model_class.query()
            }
        log.info(model_class_to_instances)
    except:
        pass


@pytest.fixture
def vcr_live_sleep(request):
    return (time.sleep
            if (request.config.getoption('record') or
                request.config.getoption('skip_vcrpy'))
            else mock.Mock())



import datetime
import logging

from lxml import html
import simplejson

from . import details
from . import essay
from . import helpers
from . import looking_for
from . import util
from .question import QuestionFetcher
from .xpath import xpb


log = logging.getLogger(__name__)


class Profile(object):
    """Represent the profile of an okcupid user.

    Many of the attributes on this object are
    :class:`~okcupyd.util.cached_property` instances which lazily load their
    values, and cache them once they have been accessed. This makes it so
    that this object avoids making unnecessary HTTP requests to retrieve the
    same piece of information twice.

    Because of this caching behavior, care must
    be taken to invalidate cached attributes on the object if an up to date view
    of the profile is needed. It is recommended that you call :meth:`.refresh`
    to accomplish this, but it is also possible to use
    :meth:`~okcupyd.util.cached_property.bust_self` to bust individual
    properties if necessary.
    """

    def __init__(self, session, username, **kwargs):
        """
        :param session: A logged in :class:`~okcupyd.session.Session`
        :param username: The username associated with the profile.
        """
        self._session = session
        #: The username of the user to whom this profile belongs.
        self.username = username
        #: A :class:`~okcupyd.util.fetchable.Fetchable` of
        #: :class:`~okcupyd.question.Question` instances, each corresponding
        #: to a question that has been answered by the user to whom this
        #: profile belongs.
        #: The fetchable consists of :class:`~okcupyd.question.UserQuestion`
        #: instead when the profile belongs to the logged in user.
        self.questions = self.question_fetchable()
        #: A :class:`~okcupyd.details.Details` instance belonging to the same
        #: user that this profile belongs to.
        self.details = details.Details(self)

        if kwargs:
            self._set_cached_properties(kwargs)

    def _set_cached_properties(self, values):
        property_names = set(
            name for name, _ in util.cached_property.get_cached_properties(self)
        )
        for key, value in values.items():
            if key not in property_names:
                log.warning("Unrecognized kwarg {0} with value {1} "
                            "passed to Profile constructor.")
            self.__dict__[key] = value

    def refresh(self, reload=False):
        """
        :param reload: Make the request to return a new profile tree. This will
                       result in the caching of the profile_tree attribute. The
                       new profile_tree will be returned.
        """
        util.cached_property.bust_caches(self, excludes=('authcode'))
        self.questions = self.question_fetchable()
        if reload:
            return self.profile_tree

    @property
    def is_logged_in_user(self):
        """
        :returns: `True` if this profile and the session it was created with
                   belong to the same user and False otherwise."""
        return self._session.log_in_name.lower() == self.username.lower()

    @util.cached_property
    def _profile_response(self):
        return self._session.okc_get(
            u'profile/{0}'.format(self.username)
        ).content

    @util.cached_property
    def profile_tree(self):
        """
        :returns: a :class:`lxml.etree` created from the html of the profile
                  page of the account associated with the username that this
                  profile was insantiated with.
        """
        return html.fromstring(self._profile_response)

    def message_request_parameters(self, content, thread_id):
        return {
            'ajax': 1,
            'sendmsg': 1,
            'r1': self.username,
            'body': content,
            'threadid': thread_id,
            'authcode': self.authcode,
            'reply': 1 if thread_id else 0,
            'from_profile': 1
        }

    @util.cached_property
    def authcode(self):
        return helpers.get_authcode(self.profile_tree)

    _photo_info_xpb = xpb.div.with_class('photo').img.select_attribute_('src')
    @util.cached_property
    def photo_infos(self):
        """
        :returns: list of :class:`~okcupyd.photo.Info` instances for each photo
                  displayed on okcupid.
        """
        from . import photo
        pics_request = self._session.okc_get(
            u'profile/{0}/album/0'.format(self.username),
        )
        pics_tree = html.fromstring(u'{0}{1}{2}'.format(
            u'<div>', pics_request.json()['fulls'], u'</div>'
        ))
        return [photo.Info.from_cdn_uri(uri)
                for uri in self._photo_info_xpb.apply_(pics_tree)]


    @util.cached_property
    def looking_for(self):
        """
        :returns: A :class:`~okcupyd.looking_for.LookingFor` instance associated
                  with this profile.
        """
        return looking_for.LookingFor(self)

    _liked_xpb = xpb.button.with_class('binary_rating_button')

    @property
    def rating(self):
        """
        Deprecated. Use :meth:`.liked` instead.

        :returns: the rating that the logged in user has given this user or
                  0 if no rating has been given.
        """
        return 5 if self.liked else 0

    @util.cached_property
    def liked(self):
        """
        :returns: Whether or not the logged in user liked this profile
        """
        if self.is_logged_in_user: return False
        classes = self._liked_xpb.one_(self.profile_tree).attrib['class'].split()
        return 'liked' in classes

    _contacted_xpb = xpb.div.with_class('actions2015').button.\
                     with_classes('actions2015-chat', 'flatbutton', 'blue').\
                     select_attribute_('data-tooltip')

    @util.cached_property
    def contacted(self):
        """
        :retuns: A boolean indicating whether the logged in user has contacted
                 the owner of this profile.
        """
        try:
            contacted_span = self._contacted_xpb.one_(self.profile_tree)
        except:
            return False
        else:
            timestamp = contacted_span.replace('Last contacted ', '')
            return helpers.parse_date_updated(timestamp)

    @util.cached_property
    def responds(self):
        """
        :returns: The frequency with which the user associated with this profile
                  responds to messages.
        """
        contacted_text = self._contacted_xpb.\
                         get_text_(self.profile_tree).lower()
        if 'contacted' not in contacted_text:
            return contacted_text.strip().replace('replies ', '')

    _id_xpb = xpb.button.with_class('binary_rating_button').\
              select_attribute_("data-tuid")

    @util.cached_property
    def id(self):
        """
        :returns: The id that okcupid.com associates with this profile.
        """
        if self.is_logged_in_user: return self._current_user_id
        return int(self._id_xpb.one_(self.profile_tree))

    @util.cached_property
    def _current_user_id(self):
        return int(helpers.get_id(self.profile_tree))

    @util.cached_property
    def essays(self):
        """
        :returns: A :class:`~okcupyd.essay.Essays` instance that is
                  associated with this profile.
        """
        return essay.Essays(self)

    _age_xpb = xpb.span.with_class('userinfo2015-basics-asl-age')
    _user_age_xpb = xpb.span(id='ajax_age')

    @util.cached_property
    def age(self):
        """
        :returns: The age of the user associated with this profile.
        """
        if self.is_logged_in_user: 
            # Retrieve the logged-in user's profile age
            return int(self._user_age_xpb.get_text_(self.profile_tree).strip())
        else:
            # Retrieve a non logged-in user's profile age
            return int(self._age_xpb.get_text_(self.profile_tree))

    _percentages_and_ratings_xpb = xpb.div.with_class('matchanalysis2015-graphs')

    @util.cached_property
    def match_percentage(self):
        """
        :returns: The match percentage of the logged in user and the user
                  associated with this object.
        """
        return int(self._percentages_and_ratings_xpb.
                   div.with_class('matchgraph--match').
                   div.with_class('matchgraph-graph').
                   canvas.select_attribute_('data-pct').
                   one_(self.profile_tree))

    @util.cached_property
    def enemy_percentage(self):
        """
        :returns: The enemy percentage of the logged in user and the user
                  associated with this object.
        """
        return int(self._percentages_and_ratings_xpb.
                   div.with_class('matchgraph--enemy').
                   div.with_class('matchgraph-graph').
                   canvas.select_attribute_('data-pct').
                   one_(self.profile_tree))

    _location_xpb = xpb.span.with_class('userinfo2015-basics-asl-location')
    _user_location_xpb = xpb.span(id='ajax_location')

    @util.cached_property
    def location(self):
        """
        :returns: The location of the user associated with this profile.
        """
        if self.is_logged_in_user: 
            # Retrieve the logged-in user's profile location
            return self._user_location_xpb.get_text_(self.profile_tree)
        else:
            # Retrieve a non logged-in user's profile location
            return self._location_xpb.get_text_(self.profile_tree)

    @util.cached_property
    def gender(self):
        """The gender of the user associated with this profile."""
        return xpb.span.with_class('ajax_gender').get_text_(self.profile_tree)

    @util.cached_property
    def orientation(self):
        """The sexual orientation of the user associated with this profile."""
        return xpb.dd(id='ajax_orientation').\
            get_text_(self.profile_tree).strip()

    @util.curry
    def message(self, message, thread_id=None):
        """Message the user associated with this profile.

        :param message: The message to send to this user.
        :param thread_id: The id of the thread to respond to, if any.
        """
        return_value = helpers.Messager(self._session).send(
            self.username, message, self.authcode, thread_id
        )
        self.refresh(reload=False)
        return return_value

    @util.cached_property
    def attractiveness(self):
        """
        :returns: The average attractiveness rating given to this profile by the
                  okcupid.com community.
        """
        # This has to be here to avoid a circular import for now.
        from .attractiveness_finder import AttractivenessFinder
        return AttractivenessFinder(self._session)(self.username)

    def toggle_like(self):
        """Toggle whether or not the logged in user likes this profile."""
        return self.rate(1 if self.liked else 5)

    def like(self):
        """Like this profile."""
        return self.rate(5)

    def unlike(self):
        """Unlike this profile."""
        return self.rate(1)

    def rate(self, rating):
        """Rate this profile as the user that was logged in with the session
        that this object was instantiated with.

        :param rating: The rating to give this user.
        """
        parameters = {
            'voterid': self._current_user_id,
            'target_userid': self.id,
            'type': 'vote',
            'cf': 'profile2',
            'target_objectid': 0,
            'vote_type': 'personality',
            'score': rating,
        }
        response = self._session.okc_post('vote_handler',
                                          data=parameters)
        response_json = response.json()
        log_function = log.info if response_json.get('status', False) \
                       else log.error
        log_function(simplejson.dumps({'rate_response': response_json,
                                       'sent_parameters': parameters,
                                       'headers': dict(self._session.headers)}))
        self.refresh(reload=False)

    def find_question(self, question_id, question_fetchable=None):
        """
        :param question_id: The id of the question to search for
        :param question_fetchable: The question fetchable to iterate through
                                   if none is provided `self.questions`
                                   will be used.
        """
        question_fetchable = question_fetchable or self.questions
        for question in question_fetchable:
            if int(question.id) == int(question_id):
                return question

    def question_fetchable(self, **kwargs):
        """
        :returns: A :class:`~okcupyd.util.fetchable.Fetchable` instance that
                  contains objects representing the answers that the user
                  associated with this profile has given to okcupid.com match
                  questions.
        """
        return util.Fetchable(QuestionFetcher(
            self._session, self.username,
            is_user=self.is_logged_in_user, **kwargs
        ))

    def authcode_get(self, path, **kwargs):
        """Perform an HTTP GET to okcupid.com using this profiles session
        where the authcode is automatically added as a query parameter.
        """
        kwargs.setdefault('params', {})['authcode'] = self.authcode
        return self._session.okc_get(path, **kwargs)

    def authcode_post(self, path, **kwargs):
        """Perform an HTTP POST to okcupid.com using this profiles session
        where the authcode is automatically added as a form item.
        """
        kwargs.setdefault('data', {})['authcode'] = self.authcode
        return self._session.okc_post(path, **kwargs)

    def __eq__(self, other):
        self.username.lower() == other.username.lower()

    def __repr__(self):
        return u'{0}("{1}")'.format(type(self).__name__, self.username)



import inspect
import itertools

import six

from . import util
from . import magicnumbers


def all_not_none_decider(function, incoming, accepted_keys):
    return all(incoming.get(key) is not None for key in accepted_keys)


def always_decider(*args, **kwargs):
    return True


class Filters(object):
    """Registrar for functions that construct filters for submission in
    requests to okcupid.com
    """

    def __init__(self, strict=True):
        self.builders = []
        self.keys = set()
        self._key_to_type = {}
        self._key_to_values = {}
        self._key_to_string = {}
        self._strict = strict

    @util.cached_property
    def filter_meta(filters_instance):
        class FilterMeta(util.decorate_all(staticmethod)):

            acceptable_values = None
            keys = ()
            types = None
            descriptions = None
            output_key = None

            def __init__(cls, name, bases, attributes_dict):
                super(FilterMeta, cls).__init__(name, bases, attributes_dict)
                cls._set_defaults()
                filters_instance.register_builder(cls)

            def decide(cls, kwargs):
                return all_not_none_decider(cls.transform, kwargs, cls.keys)

            def transform_from_kwargs(cls, kwargs):
                return cls.transform(*[kwargs.get(key) for key in cls.keys])

            def _set_defaults(cls):
                if isinstance(cls.keys, six.string_types):
                    cls.keys = [cls.keys]
                if not hasattr(cls, 'transform'):
                    cls.transform = staticmethod(lambda x: x)
                function_arguments = inspect.getargspec(cls.transform).args
                if cls.keys:
                    assert len(cls.keys) == len(function_arguments)
                else:
                    cls.keys = function_arguments

                if not cls.output_key:
                    cls.output_key = cls.keys[0]
        return FilterMeta

    @util.cached_property
    def filter_class(self):
        return six.with_metaclass(self.filter_meta)

    def build_documentation_lines(self):
        """Build a parameter documentation string that can appended to the
        docstring of a function that uses this :class:`~.Filters` instance
        to build filters.
        """
        return [
            line_string for key in sorted(self.keys)
            for line_string in self.build_paramter_string(key)
        ]

    def build_paramter_string(self, key):
        description_string = u''
        if key in self._key_to_string:
            description_string = u' {0}'.format(self._key_to_string[key])
        if key in self._key_to_values:
            description_string += u' expected values: {0}'.format(
                u', '.join([repr(value)
                           for value in self._key_to_values[key]])
            )
        parameter_string_lines = [u':param {0}:{1}'.format(
            key, description_string
        )]
        if key in self._key_to_type:
            the_type = self._key_to_type[key]
            parameter_string_lines.append(u':type {0}: {1}'.format(
                key, the_type.__name__ if isinstance(the_type, type) else the_type
            ))
        return parameter_string_lines

    def add_to_docstring_of(self, target):
        target.__doc__ = '\n    '.join(
            itertools.chain(
                (target.__doc__,), self.build_documentation_lines()
            )
        )

    all_not_none_decider = staticmethod(all_not_none_decider)

    @staticmethod
    def any_decider(function, incoming, accepted_keys):
        return bool(set(incoming).intersection(accepted_keys))

    @staticmethod
    def all_decider(function, incoming, accepted_keys):
        return set(accepted_keys).issubset(set(incoming))

    @staticmethod
    def any_not_none_decider(function, incoming, accepted_keys):
        return any(incoming.get(key) is not None for key in accepted_keys)

    def __repr__(self):
        return "{0}({1})".format(type(self).__name__,
                                 repr(self.builder_to_keys.keys()))

    def filters(self, **kwargs):
        builders = [
            builder for builder in self.builders
            if self._handle_decide(builder, kwargs)
        ]
        return [
            builder.transform(
                *[kwargs.get(key) for key in builder.keys]
            )
            for builder in builders
        ]

    def _handle_decide(self, builder, kwargs):
        if len(inspect.getargspec(builder.decide).args) == 2:
            return builder.decide(kwargs)
        else:
            return builder.decide(builder.transform, kwargs, builder.keys)

    def _validate_incoming(self, kwargs):
        if self._strict and not self.keys.issuperset(kwargs.keys()):
            raise TypeError("build() got unexpected keyword arguments: "
                            "{0}".format(', '.join(
                                repr(k) for k in kwargs.keys()
                                if k not in self.keys
                            )))

    def build(self, **kwargs):
        self._validate_incoming(kwargs)
        return {
            builder.output_key: builder.transform_from_kwargs(kwargs)
            for builder in self.builders
            if builder.decide(kwargs)
        }

    def legacy_build(self, **kwargs):
        self._validate_incoming(kwargs)
        return {
            u'filter{0}'.format(filter_number): filter_string
            for filter_number, filter_string
            in enumerate(self.filters(**kwargs), 1)
        }

    def register_builder(self, filter_object):
        self.builders.append(filter_object)
        self.keys.update(filter_object.keys)
        self._update_docs_dict(
            self._key_to_type,
            filter_object.types, filter_object.keys
        )
        self._update_docs_dict(
            self._key_to_string,
            filter_object.descriptions, filter_object.keys
        )
        self._update_docs_dict(
            self._key_to_values,
            filter_object.acceptable_values, filter_object.keys
        )

    @util.curry
    def register_filter_builder(self, function, **kwargs):
        """Register a filter function with this :class:`~.Filters` instance.
        This function is curried with :class:`~okcupyd.util.currying.curry`
        -- that is, it can be invoked partially before it is fully evaluated.
        This allows us to pass kwargs to this function when it is used as a
        decorator:

        .. code-block:: python

            @register_filter_builder(keys=('real_name',),
                                     decider=Filters.any_decider)
            def my_filter_function(argument):
                return '4,{0}'.format(argument)

        :param function: The filter function to register.
        :param keys: Keys that should be used as the argument names for
                     `function`, if none are provided, the filter functions
                     argument names will be used instead.
        :param decider: a function of signature
                        `(function, incoming_keys, accepted_keys)` that returns
                        True if the filter function should be called and False
                        otherwise. Defaults to
                        :meth:`~.all_not_none_decider`
        :param acceptable_values: A list of acceptable values for the parameter
                                  of the filter function (or a list of lists if
                                  the filter function takes multiple parameters)
        :param types: The type of the parameter accepted by the incoming filter
                      function (or a list of types if the function takes
                      multiple parameters)
        :param descriptions: A description for the incoming filter function's
                             argument (or a list of descriptions if the filter
                             function takes multiple arguments)
        :param output_key: The key to use to output the provided value. Will
                           default to the only value in keys if keys has
                           length 1.
        """
        kwargs['transform'] = function
        if kwargs.get('decider'):
            kwargs['decide'] = kwargs.get('decider')
        return type('filter', (self.filter_class,), kwargs)

    def _update_docs_dict(self, docs_dict, incoming, keys):
        if incoming:
            if not isinstance(incoming, dict):
                if len(keys) > 1:
                    assert len(keys) == len(incoming), (
                        "Got {0}, for keys: {1}".format(incoming, keys)
                    )
                    incoming = zip(keys, incoming)
                else:
                    incoming = {key: incoming for key in keys}
            docs_dict.update(incoming)


def gentation_filter(gentation):
    return u'0,{0}'.format(
        magicnumbers.gentation_to_number[gentation.strip().lower()]
    )


def age_filter(age_min=18, age_max=99):
    if age_min == None:
        age_min = 18
    return u'2,{0},{1}'.format(age_min, age_max)


def location_filter(radius):
    return u'3,{0}'.format(radius)



import logging

from okcupyd.db import model, txn, with_txn


log = logging.getLogger(__name__)


class UserAdapter(object):

    def __init__(self, profile):
        self.profile = profile

    def build(self, session):
        found = model.User.query_no_txn(session, model.User.handle ==
                                        self.profile.username)
        if found:
            return found[0]
        else:
            return model.User(okc_id=self.profile.id,
                              handle=self.profile.username,
                              age=self.profile.age,
                              location=self.profile.location)

    def get_no_txn(self, session):
        return model.User.upsert_one_no_txn(session, self.build(session),
                                            id_key='okc_id')

    get = with_txn(get_no_txn)


class ThreadAdapter(object):

    def __init__(self, thread):
        self.thread = thread

    def _get_thread(self, session):
        initiator = UserAdapter(self.thread.initiator).get_no_txn(session)
        respondent = UserAdapter(self.thread.respondent).get_no_txn(session)
        message_thread = model.MessageThread(okc_id=self.thread.id,
                                             initiator=initiator,
                                             respondent=respondent)
        return model.MessageThread.upsert_one_no_txn(session, message_thread,
                                                     id_key='okc_id')

    def _add_messages(self, thread_model):
        # This should probably use upsert.
        existing_message_ids = set([m.okc_id for m in thread_model.messages])
        new_messages = [message for message in self.thread.messages
                        if message.id not in existing_message_ids]
        new_message_models = []
        for new_message in new_messages:
            from_initiator = thread_model.initiator.handle.lower() == \
                             new_message.sender.username.lower()
            sender, recipient = (thread_model.initiator,
                                 thread_model.respondent) \
                                if from_initiator else \
                                (thread_model.respondent,
                                 thread_model.initiator)
            new_message_model = model.Message(okc_id=new_message.id,
                                              text=new_message.content,
                                              sender=sender,
                                              recipient=recipient,
                                              time_sent=new_message.time_sent)
            new_message_models.append(new_message_model)
            thread_model.messages.append(new_message_model)
        return new_message_models

    def add_messages(self):
        with txn() as session:
            thread_model = model.MessageThread.find_no_txn(session,
                                                           self.thread.id,
                                                           id_key='okc_id')
            return self._add_messages(thread_model)

    def get_thread(self):
        with txn() as session:
            thread_model = self._get_thread(session)
            return thread_model, self._add_messages(thread_model)



import logging

from lxml import html
from requests import exceptions
import simplejson

from . import errors
from . import helpers
from . import util
from .xpath import XPathBuilder, xpb


log = logging.getLogger(__name__)
thread_element_xpath = XPathBuilder(relative=False).li.with_classes('thread',
                                                                    'message')


def ThreadFetcher(session, mailbox_number):
    return util.FetchMarshall(
        ThreadHTMLFetcher(session, mailbox_number),
        util.SimpleProcessor(
            session, lambda elem: MessageThread(session, elem),
            thread_element_xpath
        )
    )


class ThreadHTMLFetcher(object):

    def __init__(self, session, mailbox_number):
        self._session = session
        self._mailbox_number = mailbox_number

    def _query_params(self, start_at):
        return {
            'low': start_at,
            'folder': self._mailbox_number,
            'infiniscroll': 1
        }

    def fetch(self, start_at):
        response = self._session.okc_get('messages',
                                         params=self._query_params(start_at))
        return response.content.strip()

    def __repr__(self):
        return '{0}(mailbox_number={1})'.format(type(self).__name__,
                                                self._mailbox_number)


class MessageFetcher(object):

    def __init__(self, session, message_thread, read_messages=False):
        self._session = session
        self._read_messages = read_messages
        self._message_thread = message_thread

    @property
    def params(self):
        return {
            'readmsg': str(self._read_messages).lower(),
            'threadid': self._message_thread.id,
            'folder': 1
        }

    @util.cached_property
    def messages_tree(self):
        messages_response = self._session.okc_get('messages',
                                                  params=self.params)
        return html.fromstring(messages_response.content.decode('utf8'))

    def refresh(self):
        util.cached_property.bust_caches(self)
        return self.messages_tree

    def fetch(self):
        for message_element in self.message_elements:
            if message_element.attrib['id'] == 'compose':
                continue
            yield Message(message_element, self._message_thread)

    _message_elements_xpb = xpb.li.with_classes('to_me', 'from_me').or_

    @util.cached_property
    def message_elements(self):
        return self._message_elements_xpb.apply_(self.messages_tree)


_base_timestamp_xpb = (
    xpb.span.with_class('timestamp').span.with_class('fancydate')
)
_timestamp_xpb = _base_timestamp_xpb.text_
_em_timestamp_xpb = _base_timestamp_xpb.em.text_


class Message(object):
    """Represent a message sent on okcupid.com"""

    def __init__(self, message_element, message_thread):
        self._message_element = message_element
        self._message_thread = message_thread

    @property
    def id(self):
        """
        :returns: The id assigned to this message by okcupid.com.
        """
        return int(self._message_element.attrib['id'].split('_')[-1])

    @util.cached_property
    def sender(self):
        """
        :returns: A :class:`~okcupyd.profile.Profile` instance belonging
                  to the sender of this message.
        """
        return (self._message_thread.user_profile
                if 'from_me' in self._message_element.attrib['class']
                else self._message_thread.correspondent_profile)

    @util.cached_property
    def recipient(self):
        """
        :returns: A :class:`~okcupyd.profile.Profile` instance belonging
                  to the recipient of this message.
        """
        return (self._message_thread.correspondent_profile
                if 'from_me' in self._message_element.attrib['class']
                else self._message_thread.user_profile)

    _content_xpb = xpb.div.with_class('message_body')

    @util.cached_property
    def content(self):
        """
        :returns: The text body of the message.
        """
        # The code that follows is obviously pretty disgusting.
        # It seems like it might be impossible to completely replicate
        # the text of the original message if it has trailing whitespace
        message = self._content_xpb.one_(self._message_element)
        first_line = message.text
        if message.text[:2] == '  ':
            first_line = message.text[2:]
        else:
            log.debug("message did not have expected leading whitespace")
        subsequent_lines = ''.join([
            html.tostring(child, encoding='unicode').replace('<br>', '\n')
            for child in message.iterchildren()
        ])
        message_text = first_line + subsequent_lines
        if len(message_text) > 0 and message_text[-1] == ' ':
            message_text = message_text[:-1]
        else:
            log.debug("message did not have expected leading whitespace")

        return message_text

    @util.cached_property
    def time_sent(self):
        try:
            timestamp_text = _timestamp_xpb.one_(self._message_element)
        except IndexError:
            timestamp_text = _em_timestamp_xpb.one_(self._message_element)
        return helpers.parse_date_updated(timestamp_text)

    def __repr__(self):
        return '<{0}: {1} sent {2} "{3}{4}">'.format(
            type(self).__name__,
            self.sender.username,
            self.recipient.username,
            self.content[:10],
            '...' if len(self.content) > 10 else ''
        )


class MessageThread(object):
    """Represent a message thread between two users."""

    @classmethod
    def delete_threads(cls, session, thread_ids_or_threads, authcode=None):
        """
        :param session: A logged in :class:`~okcupyd.session.Session`.
        :param thread_ids_or_threads: A list whose members are either
                                      :class:`~.MessageThread` instances
                                      or okc_ids of message threads.
        :param authcode: Authcode to use for this request. If none is provided
                         A request to the logged in user's messages page
                         will be made to retrieve one.
        """
        thread_ids = [thread.id if isinstance(thread, cls) else thread
                      for thread in thread_ids_or_threads]
        if not authcode:
            authcode = helpers.get_authcode(html.fromstring(
                session.okc_get('messages').content
            ))
        data = {'access_token': authcode,
                'threadids': simplejson.dumps(thread_ids)}
        return session.okc_delete('apitun/messages/threads',
                                  params=data, data=data)

    def __init__(self, session, thread_element):
        self._session = session
        self._thread_element = thread_element
        self.reply = self.correspondent_profile.message(thread_id=self.id)
        self._message_fetcher = MessageFetcher(self._session, self)
        #: A :class:`~okcupyd.util.fetchable.Fetchable` of :class:`~.Message`
        #: objects.
        self.messages = util.Fetchable(self._message_fetcher)

    @util.cached_property
    def id(self):
        """
        :returns: The id assigned to this message by okcupid.com.
        """
        return self._thread_element.attrib['data-threadid']

    @util.cached_property
    def correspondent_id(self):
        """
        :returns: The id assigned to the correspondent of this message.
        """
        try:
            return int(self._thread_element.attrib['data-personid'])
        except (ValueError, KeyError):
            try:
                return int(self.correspondent_profile.id)
            except:
                pass

    _correspondent_xpb = xpb.div.with_class('inner').a.with_class('open').\
                         span.with_class('subject').text_

    @util.cached_property
    def correspondent(self):
        """
        :returns: The username of the user with whom the logged in user is
                  conversing in this :class:`~.MessageThread`.
        """
        try:
            return self._correspondent_xpb.one_(self._thread_element).strip()
        except IndexError:
            raise errors.NoCorrespondentError()

    @util.cached_property
    def read(self):
        """
        :returns: Whether or not the user has read all the messages in this
                  :class:`~.MessageThread`.
        """
        return not 'unreadMessage' in self._thread_element.attrib['class']

    @util.cached_property
    def date(self):
        return self.datetime.date()

    @util.cached_property
    def datetime(self):
        return helpers.parse_date_updated(
            _timestamp_xpb.one_(self._thread_element)
        )

    @property
    def with_deleted_user(self):
        try:
            self.correspondent_profile.id
        except exceptions.HTTPError:
            return True
        else:
            return False

    @property
    def initiator(self):
        """
        :returns: A :class:`~okcupyd.profile.Profile` instance belonging to the
                  initiator of this :class:`~.MessageThread`.
        """
        try:
            return self.messages[0].sender
        except IndexError:
            pass

    @property
    def respondent(self):
        """
        :returns: A :class:`~okcupyd.profile.Profile` instance belonging to the
                  respondent of this :class:`~.MessageThread`.
        """
        try:
            return self.messages[0].recipient
        except IndexError:
            pass

    @util.cached_property
    def correspondent_profile(self):
        """
        :returns: The :class:`~okcupyd.profile.Profile` of the user with whom
                  the logged in user is conversing in this
                  :class:`~.MessageThread`.
        """
        return self._session.get_profile(self.correspondent)

    @util.cached_property
    def user_profile(self):
        """
        :returns: A :class:`~okcupyd.profile.Profile` belonging to the logged
                  in user.
        """
        return self._session.get_current_user_profile()

    @property
    def message_count(self):
        return len(self.messages)

    @property
    def has_messages(self):
        return bool(self.messages)

    @property
    def got_response(self):
        """
        :returns: Whether or not the :class:`~.MessageThread`. has received a
                  response.
        """
        return any(message.sender != self.initiator
                   for message in self.messages)

    def delete(self):
        """Delete this thread for the logged in user."""
        return self.delete_threads(self._session, [self.id])

    def __hash__(self):
        return hash(self.id)

    def __eq__(self, other):
        return self.id == other.id

    def __repr__(self):
        return '<{0}({1}, {2})>'.format(
            type(self).__name__,
            self.user_profile.username,
            self.correspondent_profile.username
        )



# -*- coding: utf-8 -*-
import inspect
import logging
import re

from six import string_types

from . import helpers
from . import magicnumbers
from . import util
from .magicnumbers import maps
from .xpath import xpb


log = logging.getLogger(__name__)


class Detail(object):
    """Represent a detail belonging to an okcupid.com profile."""

    NO_DEFAULT = object()

    @classmethod
    def comma_separated_presenter(cls, text):
        return text.strip().split(', ')

    @classmethod
    def mapping_multi_updater(cls, mapping):
        def updater(id_name, value):
            if value is None:
                value = ()
            return {id_name: [mapping[item.lower()] for item in value]}
        return updater

    @classmethod
    def auto_indexed_updater(cls, *options):
        return cls.mapping_updater({
            option: index
            for index, option in enumerate(options, 1)
        }, default=0)

    @classmethod
    def mapping_updater(cls, mapping, id_name=None):
        return cls(id_name=id_name,
                   updater=magicnumbers.MappingUpdater(mapping))

    @staticmethod
    def default_updater(id_name, value):
        return {id_name: value}

    def __init__(self, id_name=None, presenter=None, updater=None):
        self.id_name = id_name
        self.presenter = presenter or (lambda x: None if u'\u2014' in x
                                       else helpers.replace_chars(x.strip()))
        self.updater = updater or self.default_updater

    @property
    def id_name(self):
        return self._id_name

    _doc_format = 'The {0} detail of an okcupid.com user\'s profile.'
    @id_name.setter
    def id_name(self, value):
        self._id_name = value
        self.__doc__ = self._doc_format.format(self.id_name)

    def update(self, value):
        if isinstance(value, string_types):
            value = value.lower()
        return self.updater(self.id_name, value)

    def __get__(self, details, klass):
        if details is None:
            return self
        return self.presenter(details.id_to_display_name_value.get(self.id_name, u'\u2014'))

    def __set__(self, details, value):
        details.update(self.update(value))


class DeclarativeDetail(object):

    updater = None
    presenter = None


class Details(object):
    """Represent the details belonging to an okcupid.com profile."""

    @classmethod
    def name_detail_pairs(cls):
        is_detail = lambda x: isinstance(x, Detail)
        return inspect.getmembers(cls, is_detail)

    def __init__(self, profile):
        self.profile = profile

    _profile_details_xpb = xpb.div(id='profile_details').dl

    def refresh(self):
        util.cached_property.bust_caches(self)

    @util.cached_property
    def id_to_display_name_value(self):
        output = {}
        for element in self._profile_details_xpb.apply_(
            self.profile.profile_tree
        ):
            value_element = xpb.dd.one_(element)
            if not 'id' in value_element.attrib:
                continue
            id_name = value_element.attrib['id'].replace('ajax_', '')
            value = value_element.text_content()
            output[id_name] = value
        return output

    @property
    def as_dict(self):
        return {name: getattr(self, name)
                for name, _ in self.name_detail_pairs()}

    def convert_and_update(self, data):
        klass = type(self)
        server_bound = {}
        for key, value in data.items():
            detail = getattr(klass, key)
            server_bound.update(detail.update(value))
        return self.update(server_bound)

    def update(self, data):
        log.debug(data)
        response = self.profile.authcode_post('profileedit2', data=data)
        self.profile.refresh()
        self.refresh()
        return response

    bodytype = Detail.mapping_updater(maps.bodytype)
    orientation = Detail.mapping_updater(maps.orientation)
    smokes = Detail.mapping_updater(maps.smokes, id_name='smoking')
    drugs = Detail.mapping_updater(maps.drugs)
    drinks = Detail.mapping_updater(maps.drinks, id_name='drinking')
    job = Detail.mapping_updater(maps.job)
    status = Detail.mapping_updater(maps.status)
    monogamy = Detail(id_name='monogamous', updater=lambda id_name, value: {
        'monogamous': maps.monogamy[value],
        'monogamyflex': maps.strictness[value]
    })
    children = Detail(updater=lambda id_name, value: {
        'children': maps.has_kids[value],
        'children2': maps.wants_kids[value]
    })
    education = Detail(updater=lambda id_name, value: {
        'educationstatus': maps.education_status[value],
        'educationlevel': maps.education_level[value]
    })
    pets = Detail(updater=lambda id_name, value: {
        'cats': maps.cats[value],
        'dogs': maps.dogs[value]
    })
    diet = Detail(updater=lambda id_name, value: {
        'diet': maps.diet[value],
        'dietserious': maps.diet_strictness[value]
    })
    religion = Detail(updater=lambda id_name, value: {
        'religion': maps.religion[value],
        'religionserious': maps.seriousness[value]
    })
    sign = Detail(updater=lambda id_name, value: {
        'sign': maps.sign[value],
        'sign_status': maps.importance[value]
    })
    height = Detail(updater=lambda id_name, value: {
        'centimeters': int(round(magicnumbers.parse_height_string(value)))
    })

    class ethnicities(DeclarativeDetail):

        @staticmethod
        def presenter(text):
            return [ethnicity
                    for ethnicity in Detail.comma_separated_presenter(text)
                    if any(char.isalpha() for char in ethnicity)]

        @staticmethod
        def updater(id_name, value):
            if value is None:
                value = ()
            ethnicities = [maps.ethnicities[item.lower()] for item in value]
            if len(ethnicities) < 1:
                ethnicities = 10
            return {id_name: ethnicities}

    class income(DeclarativeDetail):

        levels = list(enumerate(
            (20000, 30000, 40000, 50000, 60000, 70000,
             80000, 100000, 150000, 250000, 500000, 1000000)
        ))

        comma_sep_number = '([0-9]{1,3}(?:,[0-9]{3})*)'

        range_matcher = re.compile(u'\$?{0}-\$?{0}'.format(comma_sep_number),
                                   flags=re.UNICODE)
        lt_matcher = re.compile("less than \$?{0}".format(comma_sep_number),
                                flags=re.UNICODE)
        gt_matcher = re.compile("more than \$?{0}".format(comma_sep_number),
                                flags=re.UNICODE)

        @classmethod
        def updater(cls, id_name, value):
            if value is None:
                return {'income': 0}
            if isinstance(value, string_types):
                for matcher, sign in ((cls.range_matcher, 1),
                                      (cls.lt_matcher, -1),
                                      (cls.gt_matcher, 1)):
                    match = matcher.match(value)
                    if match:
                        matched_income = int(match.group(1).replace(',', ''))
                        value = matched_income + 100 * sign
                        break
            for index, level in cls.levels:
                if value < level:
                    break
            else:
                index += 1
            update = index + 1
            return {'income': update}

    class languages(DeclarativeDetail):

        language_matcher = re.compile('(.*?) \((.*?)\)')

        language_to_number = magicnumbers.language_map_2

        level = util.IndexedREMap('fluently', 'okay', 'poorly')

        @classmethod
        def presenter(cls, value):
            language_strings = value.split(',')
            languages = []
            for language_string in language_strings:
                match = cls.language_matcher.match(language_string.strip())
                if match:
                    languages.append((match.group(1).lower(),
                                      match.group(2).lower()))
                else:
                    languages.append((language_string.strip(), None))
            return languages

        @classmethod
        def updater(cls, id_name, languages):
            data = {}
            number = 0
            for number, (language, level) in enumerate(languages, 1):
                language_number = cls.language_to_number[language.lower()]
                level = level or ''
                level_number = cls.level[level.lower()]
                data['cont_lang_{0}'.format(number)] = language_number
                data['language{0}status'.format(number)] = level_number
            number += 1
            for i in range(number, 6):
                data['cont_lang_{0}'.format(i)] = ''
                data['language{0}status'.format(i)] = ''
            return data


for id_name, detail in Details.name_detail_pairs():
    if detail.id_name is None:
        detail.id_name = id_name

    is_declarative_detail = lambda x: (isinstance(x, type) and
                                  issubclass(x, DeclarativeDetail))
    for id_name, declarative_detail in inspect.getmembers(
        Details, is_declarative_detail
    ):
        detail = Detail(presenter=declarative_detail.presenter,
                        updater=declarative_detail.updater,
                        id_name=id_name)
        setattr(Details, id_name, detail)



from sqlalchemy.types import TypeDecorator, VARCHAR
import simplejson


class JSONType(TypeDecorator):

    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = simplejson.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = simplejson.loads(value)
        return value


class StringBackedInteger(TypeDecorator):

    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        return str(value)

    def process_result_value(self, value, dialect):
        return int(value)


"""Module where the default username and password for logging in to
okcupid are housed.
"""
import os
import yaml

#: The username that will be used to log in to okcupid
USERNAME = os.environ.get('OKC_USERNAME')
#: The password that will be used to log in to okcupid
PASSWORD = os.environ.get('OKC_PASSWORD')

AF_USERNAME = os.environ.get('AF_USERNAME', USERNAME)
AF_PASSWORD = os.environ.get('AF_PASSWORD', PASSWORD)

OKCUPYD_CONFIG_FILENAME = '.okcupyd.yml'


def okcupyd_config_at_path(path):
    return os.path.join(path, OKCUPYD_CONFIG_FILENAME)


def generate_paths_to_check():
    yield os.getcwd()
    yield os.path.expanduser("~")


def load_credentials_from_filepath(filepath):
    global USERNAME, PASSWORD
    with open(filepath, 'r') as file_object:
        data = yaml.load(file_object.read())
        USERNAME = data['username']
        PASSWORD = data['password']


def load_credentials_from_files():
    for path in generate_paths_to_check():
        filepath = okcupyd_config_at_path(path)
        if os.path.exists(filepath):
            load_credentials_from_filepath(filepath)
            return filepath


if USERNAME is None:
    load_credentials_from_files()



import itertools
import numbers

from . import util
from .attractiveness_finder import AttractivenessFinder


class Statistics(object):

    def __init__(self, user, message_threads=None, filters=(),
                 attractiveness_finder=None):
        self._user = user
        self._message_threads = message_threads or set(itertools.chain(user.inbox,
                                                                       user.outbox))
        self._filters = filters
        self._attractiveness_finder = attractiveness_finder or AttractivenessFinder()

    def _thread_matches(self, message_thread):
        return all(f(message_thread) for f in self._filters)

    @util.cached_property
    def threads(self):
        return set(mt for mt in self._message_threads if self._thread_matches(mt))

    @util.cached_property
    def has_messages(self):
        return self.with_filters(lambda mt: mt.has_messages)

    @util.cached_property
    def has_response(self):
        return self.with_filters(lambda mt: mt.got_response)

    @util.cached_property
    def no_responses(self):
        return self.with_filters(lambda mt: not mt.got_response)

    @util.cached_property
    def initiated(self):
        return self.with_filters(lambda mt: mt.initiator == self._user.profile)

    @util.cached_property
    def received(self):
        return self.with_filters(lambda mt: mt.initiator != self._user.profile)

    @util.cached_property
    def has_attractiveness(self):
        return self.with_filters(lambda mt: self._attractiveness_finder.find_attractiveness(
            mt.correspondent) is not None)

    def time_filter(self, min_date=None, max_date=None):
        def _time_filter(thread):
            if min_date and min_date > thread.date:
                return False
            if max_date and max_date < thread.date:
                return False
            return True
        return self.with_filters(_time_filter)

    def attractiveness_filter(self, attractiveness_finder=None,
                                   min_attractiveness=0, max_attractiveness=10000):
        attractiveness_finder = attractiveness_finder or self._attractiveness_finder
        def _attractiveness_filter(thread):
            attractiveness = attractiveness_finder.find_attractiveness(
                thread.correspondent
            )
            return (isinstance(attractiveness, numbers.Number) and
                    min_attractiveness <= attractiveness <= max_attractiveness)
        return self.with_filters(_attractiveness_filter)

    def with_filters(self, *filters, **kwargs):
        message_threads = self.threads \
                          if kwargs.get('apply_filters_immediately', True) \
                          else self._message_threads
        return type(self)(self._user, message_threads, self._filters + filters,
                          attractiveness_finder=self._attractiveness_finder)

    @property
    def count(self):
        return len(self.threads)

    @property
    def response_rate(self):
        return float(self.has_response.count)/self.count

    def _average(self, function):
        return sum(map(function, self.threads))/self.count

    @property
    def average_first_message_length(self):
        return self._average(lambda thread: len(thread.messages[0].content))

    @property
    def average_conversation_length(self):
        return self._average(lambda thread: thread.message_count)

    def _average_attractiveness(self, attractiveness_finder=None):
        attractiveness_finder = attractiveness_finder or self._attractiveness_finder
        return self.has_attractiveness._average(lambda thread: (
            attractiveness_finder.find_attractiveness(
                thread.correspondent
            ))
        )

    @property
    def average_attractiveness(self):
        return self._average_attractiveness()

    @property
    def portion_initiated(self):
        return self.initiated.count/self.count

    @property
    def portion_received(self):
        return 1 - self.portion_initiated



import logging

import simplejson
import six

from . import filter
from . import helpers
from . import magicnumbers
from . import util
from .profile import Profile
from .session import Session


log = logging.getLogger(__name__)


search_filters = filter.Filters(strict=False)


# The docstring below is extended automatically. Read it in its entirety at
# http://okcupyd.readthedocs.org/en/latest/ or by generating the documentation
# yourself.
def SearchFetchable(session=None, **kwargs):
    """Search okcupid.com with the given parameters. Parameters are
    registered to this function through
    :meth:`~okcupyd.filter.Filters.register_filter_builder` of
    :data:`~okcupyd.json_search.search_filters`.

    :returns: A :class:`~okcupyd.util.fetchable.Fetchable` of
              :class:`~okcupyd.profile.Profile` instances.

    :param session: A logged in session.
    :type session: :class:`~okcupyd.session.Session`
    """
    session = session or Session.login()
    return util.Fetchable(
        SearchManager(
            SearchJSONFetcher(session, **kwargs),
            ProfileBuilder(session)
        )
    )


class SearchManager(object):

    def __init__(self, search_fetchable, profile_builder):
        self._search_fetchable = search_fetchable
        self._profile_builder = profile_builder
        self._last_after = None

    def fetch(self, count=18):
        last_last_after = object()
        while last_last_after != self._last_after:
            last_last_after = self._last_after
            for profile in self.fetch_once(count=count):
                yield profile

    def fetch_once(self, count=18):
        response = self._search_fetchable.fetch(
            after=self._last_after, count=count
        )
        try:
            self._last_after = response['paging']['cursors']['after']
        except KeyError:
            log.warning(simplejson.dumps(
                {
                    'msg': "unable to get after cursor from response",
                    'response': response
                }
            ))
        for profile in self._profile_builder(response):
            yield profile


class SearchJSONFetcher(object):

    search_uri = '1/apitun/match/search'
    default_headers = {
        'Content-Type': 'application/json'
    }

    def __init__(self, session=None, **options):
        self._session = session or Session.login()
        self._options = options
        self._parameters = search_filters.build(session=self._session, **options)

    def _get_headers(self):
        headers = helpers.make_auth_header_dictionary(self._session)
        headers.update(self.default_headers)
        return headers

    def _request_params(self, after=None, count=None):
        return {
            'headers': self._get_headers(),
            'data': simplejson.dumps(self._post_body(after, count)),
            'path': self.search_uri,
        }

    def _post_body(self, after=None, count=None):
        search_parameters = {
            'after': after,
            'limit': count,
            'fields': "userinfo,thumbs,percentages,likes,last_contacts,online",
        }
        search_parameters.update(self._parameters)
        return search_parameters

    def fetch(self, after=None, count=18):
        request_parameters = self._request_params(after=after, count=count)
        log.info(simplejson.dumps(request_parameters))
        response = self._session.okc_post(**request_parameters)
        try:
            search_json = response.json()
        except:
            log.warning(simplejson.dumps({'failure': response.content}))
            raise
        return search_json


class ProfileBuilder(object):

    def __init__(self, session):
        self._session = session

    def __call__(self, response_dictionary):
        try:
            profile_infos = response_dictionary['data']
        except KeyError:
            log.warning(simplejson.dumps(
                {
                    'msg': "unable to get data from response",
                    'response': response_dictionary
                }
            ))
        else:
            for profile_info in profile_infos:
                yield Profile(self._session, profile_info["username"])


class GentationFilter(search_filters.filter_class):

    def transform(gentation):
        if isinstance(gentation, six.string_types):
            gentations = [gentation]
        else:
            gentations = gentation
        return [
            magicnumbers.gentation_to_number.get(
                a_gentation.strip().lower(), gentation
            )
            for a_gentation in gentations
        ]

    descriptions = "A list of the allowable gentations of returned search results."
    types = list
    acceptable_values = magicnumbers.gentation_to_number.keys()


class MinimumAgeFilter(search_filters.filter_class):

    keys = 'minimum_age'

    descriptions = "Filter profiles with ages above the provided value."
    types = int


class MaximumAgeFilter(search_filters.filter_class):

    keys = 'maximum_age'

    descriptions = "Filter profiles with ages below the provided value."
    types = int


class RadiusFilter(search_filters.filter_class):

    output_key = "radius"
    descriptions = "The maximum distance (in miles) from the specified location of returned search results."
    types = "int or None"

    def transform(radius):
        return radius


class LocationActivationFilter(search_filters.filter_class):

    output_key = "located_anywhere"
    types = (int, type(None))

    def transform(radius):
        return 1 if radius is None else 0


class LocIdFilter(search_filters.filter_class):

    types = int

    def transform(locid):
        return locid


class LocationFilter(search_filters.filter_class):

    output_key = "locid"
    descriptions = (
        "A query that will be used to look up a locid for the search, location_cache must also be passed in in order for this parameter to work. :class:`~okcupyd.user.User` automatically passes the location_cache in.",
        "A :class:`~okcupyd.location.LocationQueryCache` instance."
    )

    # TODO:
    # Requiring location_cache here is a weird type of dependency
    # injection. A better solution is probably needed long term
    def transform(location, location_cache):
        return location_cache.get_locid(location)


def search(session=None, count=1, **kwargs):
    return SearchFetchable(session, count=count, **kwargs)[:count]


class OrderByFilter(search_filters.filter_class):

    def transform(order_by):
        return (order_by or "SPECIAL_BLEND").upper()


search_filters.add_to_docstring_of(SearchFetchable)



from .util import cached_property


class XPathBuilder(object):

    def __init__(self, nodes=(), relative=True, direct_child=False):
        self.nodes = tuple(nodes)
        self.relative = relative
        self.direct_child = direct_child

    @cached_property
    def xpath(self):
        return ('.' if self.relative else '') + ''.join(node.xpath
                                                        for node in self.nodes)

    @property
    def or_(self):
        return self.update_final_node(self.nodes[-1].make_or)

    @property
    def text_(self):
        return self.update_final_node(
            self.nodes[-1](selected_attribute=XPathNode.text)
        )

    def add_node(self, **kwargs):
        if 'direct_child' not in kwargs:
            kwargs['direct_child'] = self.direct_child
        return type(self)(self.nodes + (XPathNode(**kwargs),),
                          relative=self.relative)

    def __getattr__(self, attr):
        return self.add_node(element=attr)

    def update_final_node(self, updated_final_node):
        return type(self)(self.nodes[:-1] + (updated_final_node,),
                          relative=self.relative,
                          direct_child=self.direct_child)

    def __call__(self, *predicates, **attributes):
        direct_child = attributes.pop('direct_child', None)
        assert len(self.nodes)
        updated_final_node = self.nodes[-1](predicates=predicates,
                                            attributes=attributes,
                                            direct_child=direct_child)
        return self.update_final_node(updated_final_node)

    def attribute_contains(self, attribute, contains_string):
        updated_final_node = self.nodes[-1].add_contains_predicates(
            ((attribute, contains_string),)
        )
        return self.update_final_node(updated_final_node)

    def with_classes(self, *classes):
        return self.update_final_node(self.nodes[-1].with_classes(classes))

    def select_attribute_(self, attribute, elem=None):
        update_final_node = self.nodes[-1](selected_attribute=attribute)
        builder = self.update_final_node(update_final_node)
        if elem is not None:
            return builder.apply_(elem)
        else:
            return builder

    def text_contains_(self, contained_text):
        updated_final_node = self.nodes[-1].text_contains(contained_text)
        return self.update_final_node(updated_final_node)

    with_class = with_classes

    def apply_(self, tree):
        return tree.xpath(self.xpath)

    def one_(self, tree):
        return self.apply_(tree)[0]

    def get_text_(self, tree):
        return self.apply_(tree)[0].text_content()

    def __repr__(self):
        return '{0}("{1}")'.format(type(self).__name__, self.xpath)


class XPathNode(object):

    text = object()

    @staticmethod
    def contains_class(class_attribute, contained_class):
        return "contains(concat(' ',normalize-space(@{0}),' '),' {1} ')".\
            format(class_attribute, contained_class)

    @staticmethod
    def contains_attribute(attribute, contained_string):
        return "contains(@{0}, '{1}')".format(attribute, contained_string)

    @staticmethod
    def attribute_equal(attribute, value):
        return "@{0} = '{1}'".format(attribute, value)

    def __init__(self, element='*', attributes=None, predicates=None,
                 direct_child=False, use_or=False, selected_attribute=None):
        self.element = element
        self.predicates = tuple(predicates) if predicates else ()
        if attributes:
            self.predicates += tuple([self.attribute_equal(key, value)
                                      for key, value in attributes.items()])
        self.direct_child = direct_child
        self.use_or = use_or
        self.selected_attribute = selected_attribute

    @property
    def make_or(self):
        return self(use_or=True)

    @property
    def separator(self):
        return '/' if self.direct_child else '//'

    @property
    def xpath(self):
        return '{0}{1}{2}{3}'.format(self.separator, self.element,
                                     self.predicate_string,
                                     self.selected_attribute_string)

    @property
    def predicate_joiner(self):
        return ' or ' if self.use_or else ' and '

    @property
    def predicate_string(self):
        if self.predicates:
            predicate = self.predicate_joiner.join(self.predicates)
            return '[ {0} ]'.format(predicate)
        else:
            return ''

    @property
    def selected_attribute_string(self):
        if self.selected_attribute is self.text:
            return '/text()'
        return '/@{0}'.format(self.selected_attribute) \
            if self.selected_attribute else ''

    def __call__(self, element=None, predicates=(), attributes=None,
                 direct_child=None, use_or=False, selected_attribute=None):
        direct_child = (self.direct_child
                        if direct_child is None
                        else direct_child)
        element = self.element if element is None else element
        new_predicates = self.predicates + tuple(predicates)
        return type(self)(element, attributes, new_predicates,
                          direct_child, use_or, selected_attribute)

    def with_classes(self, classes):
        predicates = tuple(self.contains_class('class', contained_class)
                           for contained_class in classes)

        return self(predicates=predicates)

    def add_contains_predicates(self, kv_pairs):
        predicates = tuple(self.contains_attribute(attribute, contains_string)
                           for attribute, contains_string in kv_pairs)
        return self(predicates=predicates)

    def text_contains(self, contained_text):
        return self(predicates=("contains(text(),'{0}')".
                                format(contained_text),))


xpb = XPathBuilder()



import logging

import simplejson

from okcupyd import helpers
from okcupyd.db import adapters
from okcupyd.db import model, txn
from okcupyd.util import curry


log = logging.getLogger(__name__)


class Sync(object):
    """Sync messages from a users inbox to the okc database."""

    def __init__(self, user):
        self._user = user

    def all(self):
        self.update_mailbox('outbox')
        return self.update_mailbox('inbox')

    @curry
    def update_mailbox(self, mailbox_name='inbox'):
        """Update the mailbox associated with the given mailbox name.
        """
        with txn() as session:
            last_updated_name = '{0}_last_updated'.format(mailbox_name)
            okcupyd_user = session.query(model.OKCupydUser).join(model.User).filter(
                model.User.okc_id == self._user.profile.id
            ).with_for_update().one()
            log.info(simplejson.dumps({
                '{0}_last_updated'.format(mailbox_name): helpers.datetime_to_string(
                    getattr(okcupyd_user, last_updated_name)
                )
            }))

            res = self._sync_mailbox_until(
                getattr(self._user, mailbox_name)(),
                getattr(okcupyd_user, last_updated_name)
            )
            if not res:
                return None, None

            last_updated, threads, new_messages = res
            if last_updated:
                setattr(okcupyd_user, last_updated_name, last_updated)
            return threads, new_messages

    inbox = update_mailbox(mailbox_name='inbox')
    outbox = update_mailbox(mailbox_name='outbox')

    def _sync_mailbox_until(self, mailbox, sync_until):
        threads = []
        messages = []
        for thread in mailbox:
            if sync_until and sync_until > thread.datetime:
                break
            if not thread.messages:
                continue
            if not thread.with_deleted_user:
                thread, new_messages = adapters.ThreadAdapter(thread).get_thread()
                threads.append(threads)
                messages.extend(new_messages)
        try:
            return mailbox[0].datetime, threads, messages
        except IndexError:
            pass



import importlib
import logging

from invoke import Collection, task

from okcupyd.user import User
from okcupyd.profile_copy import Copy


log = logging.getLogger(__name__)
ns = Collection()


def build_copy(source_credentials_or_username, dest_credentials):
    dest_module = importlib.import_module(dest_credentials)
    dest_user = User.from_credentials(dest_module.USERNAME,
                                      dest_module.PASSWORD)
    try:
        source_module = importlib.import_module(source_credentials_or_username)
    except ImportError:
        source = dest_user.get_profile(source_credentials_or_username)
    else:
        source = User.from_credentials(source_module.USERNAME,
                                       source_module.PASSWORD)

    return Copy(source, dest_user)


def build_copy_task(method):
    @task
    def copy_task(source_credentials, dest_credentials):
        copy = build_copy(source_credentials, dest_credentials)
        getattr(copy, method)()
    ns.add_task(copy_task, name=method)


for method in Copy.copy_methods:
    build_copy_task(method)


@ns.add_task
@task(default=True)
def all(source_credentials, dest_credentials):
    log.info(build_copy(source_credentials, dest_credentials).all())



import os

from setuptools import setup, find_packages


version = '1.0.0alpha6'


with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as file:
    long_description = file.read()


setup(
    name="okcupyd",
    version=version,
    packages=find_packages(exclude=('tests*', 'examples')),
    install_requires=['lxml', 'requests ~= 2.7', 'simplejson ~= 3.8',
                      'sqlalchemy ~= 1.0', 'ipython ~= 5.0.0',
                      'wrapt ~= 1.10', 'coloredlogs == 5.0', 'invoke ~= 0.13',
                      'six ~= 1.10', 'setuptools ~= 25.1.0', 'PyYAML >= 1.1'],
    tests_require=['tox', 'pytest', 'mock', 'contextlib2', 'vcrpy >= 1.7.0'],
    package_data={'': ['*.md', '*.rst']},
    author="Ivan Malison",
    author_email="ivanmalison@gmail.com",
    description="A package for interacting with okcupid.com",
    license="MIT",
    keywords=["okcupid", "okcupyd", "pyokc", "online dating"],
    url="https://github.com/IvanMalison/okcupyd",
    long_description=long_description,
    entry_points={"console_scripts": ["okcupyd=okcupyd:interactive"]},
    classifiers=[
        "Development Status :: 3 - Alpha",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Topic :: Utilities",
    ],
)



import time

from invoke import task

from okcupyd import User


@task
def answer_all_their_questions(username):
    user = User()
    them = user.get_profile(username)

    for question in them.questions:
        if not question.answered:
            user.questions.respond(question.id, [1], [1], 3)

    questions = them.question_fetchable()[:]
    print len(questions)
    question_id_to_question = {question.id: question
                               for question in user.profile.questions}
    user.questions.clear()
    time.sleep(10)
    for question in questions:
        user.questions.respond_from_question(question,
                                             question_id_to_question, 3)

@task
def rate_attractive_girls():
    user = User()
    while True:
        p = user.quickmatch()
        if user.attractiveness_finder(p.username) >= 7000:
            p.rate(5)



from . import helpers
from .session import Session


class NoLocationFoundError(Exception):
    pass


class LocationQueryCache(object):

    uri = '1/apitun/location/query'

    def __init__(self, session=None):
        self._session = session or Session.login()
        self._locid_cache = {}
        self._query_cache = {}

    def get_locid(self, query):
        if query in self._query_cache:
            return self._query_cache[query]
        result = self._query(query)
        if 'results' in result and len(result['results']):
            locid = result['results'][0]['locid']
            if locid not in self._locid_cache:
                self._locid_cache[locid] = result['results'][0]
            return locid
        raise NoLocationFoundError()

    def get(self, query):
        locid = self.get_locid(query)
        if locid:
            return self._locid_cache[locid]

    def _query(self, query):
        return self._session.okc_get(
            self.uri,
            params={'q': query},
            headers=helpers.make_auth_header_dictionary(self._session)
        ).json()



import functools
import logging

from . import util
from .xpath import xpb


log = logging.getLogger(__name__)


class BaseQuestion(object):
    """The abstract base class of :class:`~.Question` and
    :class:`~.UserQuestion`. Contains all the shared functionality of the
    aforementined classes. The two are quite different in some ways and can
    not always be used interchangably. See their respective docstrings for more
    details.
    """

    def __init__(self, question_element):
        self._question_element = question_element

    @property
    def answered(self):
        return 'not_answered' not in self._question_element.attrib['class']

    @util.cached_property
    def id(self):
        """
        :returns: The integer id given to this question by okcupid.com.
        """
        return int(self._question_element.attrib['data-qid'])

    _text_xpb = xpb.div.with_class('qtext').p

    @util.cached_property
    def text(self):
        return self._text_xpb.get_text_(self._question_element).strip()

    def __repr__(self):
        return u'<{1}: {0}>'.format(self.text, type(self).__name__)


class Question(BaseQuestion):
    """Represent a question answered by a user other than the logged in user.

    Note: Because of the way that okcupid presents question data it is actually
    not very easy to get the index of the answer to a question that belongs
    to a user other than the logged in user. It is possible to retrieve
    this value (see :meth:`okcupyd.user.User.get_question_answer_id` and
    :meth:`~.UserQuestion.get_answer_id_for_question`), but it
    can take quite a few requests to do so. For this reason, the answer_id is
    NOT included as an attribute on this object, despite its inclusion in
    :class:`~.UserQuestion`.
    """

    def return_none_if_unanswered(function):
        @functools.wraps(function)
        def wrapped(self):
            if self.answered:
                return function(self)
        return wrapped

    _answers_xpb = xpb.div.with_class('answers').\
                   p.with_class('answer')

    def __init__(self, question_element):
        super(Question, self).__init__(question_element)
        try:
            self._their_answer_span, self._my_answer_span = (
                self._answers_xpb.span.with_class('text').apply_(
                    self._question_element
                )
            )
            self._their_note_span, self._my_note_span = (
                self._answers_xpb.span.with_class('note').apply_(
                    self._question_element
                )
            )
        except:
            pass

    @util.cached_property
    @return_none_if_unanswered
    def their_answer(self):
        """The answer that the user whose :class:`~okcupyd.profile.Profile` this
        question was retrieved from provided to this
        :class:`~okcupyd.question.Question`.
        """
        return self._their_answer_span.text_content().strip()

    @util.cached_property
    @return_none_if_unanswered
    def my_answer(self):
        """The answer that the user whose :class:`~okcupyd.session.Session` was
        used to create this :class:`~okcupyd.question.Question` provided.
        """
        return self._my_answer_span.text_content().strip()

    @util.cached_property
    @return_none_if_unanswered
    def their_answer_matches(self):
        """
        :returns: whether or not the answer provided by the user
                  answering the question is acceptable to the logged in user.
        :rtype: rbool
        """
        return 'not_accepted' not in self._their_answer_span.attrib['class']

    @util.cached_property
    @return_none_if_unanswered
    def my_answer_matches(self):
        """
        :returns: bool indicating whether or not the answer provided
                  by the logged in user is acceptable to the user
                  answering the question.
        """
        return 'not_accepted' not in self._my_answer_span.attrib['class']

    @util.cached_property
    @return_none_if_unanswered
    def their_note(self):
        """
        :returns: The note the answering user provided as explanation for their
                  answer to this question.
        """
        return self._their_note_span.text_content().strip()

    @util.cached_property
    @return_none_if_unanswered
    def my_note(self):
        """
        :returns: The note the logged in user provided as an explanation for
                  their answer to this question.
        """
        return self._my_note_span.text_content().strip()

    _explanation_xpb = xpb.div.span.with_class('note')
    del return_none_if_unanswered


class UserQuestion(BaseQuestion):
    """Represent a question answered by the logged in user."""

    _answer_option_xpb = xpb.ul.with_class('self_answers').li
    _explanation_xpb = xpb.div.with_class('your_explanation').\
                       p.with_class('value')

    def get_answer_id_for_question(self, question):
        """Get the answer_id corresponding to the answer given for question
        by looking at this :class:`~.UserQuestion`'s answer_options.
        The given :class:`~.Question` instance must have the same id as this
        :class:`~.UserQuestion`.

        That this method exists is admittedly somewhat weird. Unfortunately, it
        seems to be the only way to retrieve this information.
        """
        assert question.id == self.id
        for answer_option in self.answer_options:
            if answer_option.text == question.their_answer:
                return answer_option.id

    @util.cached_property
    def answer_id(self):
        for answer_option in self.answer_options:
            if answer_option.is_users:
                return answer_option.id

    @util.cached_property
    def answer_options(self):
        """
        :returns: A list of :class:`~.AnswerOption` instances representing the
                  available answers to this question.
        """
        return [
            AnswerOption(element)
            for element in self._answer_option_xpb.apply_(
                self._question_element
            )
        ]

    @util.cached_property
    def explanation(self):
        """
        :returns: The explanation written by the logged in user for this
                  question (if any).
        """
        try:
            return self._explanation_xpb.get_text_(self._question_element)
        except:
            pass

    @util.cached_property
    def answer_text_to_option(self):
        return {option.text: option for option in self.answer_options}

    @util.cached_property
    def answer(self):
        """
        :returns: A :class:`~.AnswerOption` instance corresponding to the answer
                  the user gave to this question.
        """
        for answer_option in self.answer_options:
            if answer_option.is_users:
                return answer_option


class AnswerOption(object):

    def __init__(self, option_element):
        self._element = option_element

    @util.cached_property
    def is_users(self):
        """
        :returns: Whether or not this was the answer selected by the logged in
                  user.
        """
        return 'mine' in self._element.attrib['class']

    @util.cached_property
    def is_match(self):
        """
        :returns: Whether or not this was the answer is acceptable to the logged
                  in user.
        """
        return 'match' in self._element.attrib['class']

    @util.cached_property
    def text(self):
        """
        :returns: The text of this answer.
        """
        return self._element.text_content().strip()

    @util.cached_property
    def id(self):
        """
        :returns: The integer index associated with this answer.
        """
        return int(self._element.attrib['id'].split('_')[-1])

    def __repr__(self):
        return '<{0}: "{1}" (is_users={2}, is_match={3})>'.format(
            type(self).__name__,
            self.text,
            self.is_users,
            self.is_match
        )


importances = ('not_important', 'little_important', 'somewhat_important',
               'very_important', 'mandatory')


class Questions(object):
    """Interface to accessing and answering questions belonging to the logged
    in user."""

    headers = {
        'accept': 'application/json, text/javascript, */*; q=0.01',
        'accept-encoding': 'gzip,deflate',
        'accept-language': 'en-US,en;q=0.8',
        'content-type': 'application/x-www-form-urlencoded; charset=UTF-8',
        'origin': 'https://www.okcupid.com',
        'referer': 'https://www.okcupid.com/questions',
        'x-requested-with': 'XMLHttpRequest',
    }

    #: Human readable importance name to integer used to represent them on
    #: okcupid.com
    importance_name_to_number = {
        'mandatory': 0,
        'very_important': 1,
        'somewhat_important': 3,
        'little_important': 4,
        'not_important': 5
    }

    path = 'questions/ask'

    def __init__(self, session, importances=importances, user_id=None):
        self.importance_name_to_fetchable = {}
        for importance in importances:
            fetchable = util.Fetchable.fetch_marshall(
                QuestionHTMLFetcher(session, 'questions', **{importance: 1}),
                QuestionProcessor(UserQuestion)
            )
            self.importance_name_to_fetchable[importance] = fetchable
            setattr(self, importance, fetchable)
        self._session = session
        if user_id:
            self._user_id = user_id

    @util.cached_property
    def _user_id(self):
        return self._session.get_current_user_profile().id

    def respond_from_user_question(self, user_question, importance):
        """Respond to a question in exactly the way that is described by
        the given user_question.

        :param user_question: The user question to respond with.
        :type user_question: :class:`.UserQuestion`
        :param importance: The importance that should be used in responding to
                           the question.
        :type importance: int see :attr:`.importance_name_to_number`
        """
        user_response_ids = [option.id
                             for option in user_question.answer_options
                             if option.is_users]
        match_response_ids = [option.id
                              for option in user_question.answer_options
                              if option.is_match]
        if len(match_response_ids) == len(user_question.answer_options):
            match_response_ids = 'irrelevant'
        return self.respond(user_question.id, user_response_ids,
                            match_response_ids, importance,
                            note=user_question.explanation or '')

    def respond_from_question(self, question, user_question, importance):
        """Copy the answer given in `question` to the logged in user's
        profile.

        :param question: A :class:`~.Question` instance to copy.
        :param user_question: An instance of :class:`~.UserQuestion` that
                              corresponds to the same question as `question`.
                              This is needed to retrieve the answer id from
                              the question text answer on question.
        :param importance: The importance to assign to the response to the
                           answered question.
        """
        option_index = user_question.answer_text_to_option[
            question.their_answer
        ].id
        self.respond(question.id, [option_index], [option_index], importance)

    def respond(self, question_id, user_response_ids, match_response_ids,
                importance, note='', is_public=1, is_new=1):
        """Respond to an okcupid.com question.

        :param question_id: The okcupid id used to identify this question.
        :param user_response_ids: The answer id(s) to provide to this question.
        :param match_response_ids: The answer id(s) that the user considers
                                   acceptable.
        :param importance: The importance to attribute to this question. See
                           :attr:`.importance_name_to_number` for details.
        :param note: The explanation note to add to this question.
        :param is_public: Whether or not the question answer should be made
                          public.
        """
        form_data = {
            'ajax': 1,
            'submit': 1,
            'answer_question': 1,
            'skip': 0,
            'show_all': 0,
            'targetid': self._user_id,
            'qid': question_id,
            'answers': user_response_ids,
            'matchanswers': match_response_ids,
            'is_new': is_new,
            'is_public': is_public,
            'note': note,
            'importance': importance,
            'delete_note': 0
        }
        return self._session.okc_post(
            self.path, data=form_data, headers=self.headers
        )

    def clear(self):
        """USE WITH CAUTION. Delete the answer to every question that the
        logged in user has responded to.
        """
        return self._session.okc_post(
            'questions', data={'clear_all': 1}, headers=self.headers
        )


_page_data_xpb = xpb.div.with_class('pages_data')
_current_page_xpb = _page_data_xpb.input(id='questions_pages_page').\
                    select_attribute_('value')
_total_page_xpb = _page_data_xpb.input(id='questions_pages_total').\
                  select_attribute_('value')
_question_xpb = xpb.div.with_class('question')


def QuestionProcessor(question_class):
    return util.PaginationProcessor(question_class, _question_xpb,
                                    _current_page_xpb, _total_page_xpb)


class QuestionHTMLFetcher(object):

    @classmethod
    def from_username(cls, session, username, **kwargs):
        return cls(session, u'profile/{0}/questions'.format(username), **kwargs)

    def __init__(self, session, uri, **additional_parameters):
        self._session = session
        self._uri = uri
        self._additional_parameters = additional_parameters

    def _query_params(self, start_at):
        parameters = {'low': start_at, 'leanmode': 1}
        parameters.update(self._additional_parameters)
        return parameters

    def fetch(self, start_at):
        response = self._session.okc_get(self._uri,
                                         params=self._query_params(start_at))
        return response.content.decode('utf8', 'replace')


def QuestionFetcher(session, username, question_class=Question,
                    is_user=False, **kwargs):
    if is_user:
        question_class = UserQuestion
    return util.FetchMarshall(
        QuestionHTMLFetcher.from_username(session, username, **kwargs),
        QuestionProcessor(question_class)
    )



# -*- coding: utf-8 -*-
import itertools
import re

import six

from okcupyd import util
from okcupyd.question import UserQuestion


class maps(six.with_metaclass(util.GetAttrGetItem)):

    bodytype = util.IndexedREMap(
        'rather not say', 'thin', 'overweight', 'skinny', 'average', 'fit',
        'athletic', 'jacked', 'a little extra', 'curvy', 'full figured',
        'used up'
    )

    orientation = util.IndexedREMap('straight', 'gay', 'bisexual')

    smokes = util.IndexedREMap(
        'yes', 'sometimes', 'when drinking', 'trying to quit', 'no'
    )

    drugs = util.IndexedREMap('never', 'sometimes', 'often',
                              default=3, offset=0)

    drinks = util.IndexedREMap('very often', 'often', 'socially', 'rarely',
                               'desperately', 'not at all')

    ethnicities = util.IndexedREMap(
        'asian', 'middle eastern', 'black', 'native american', 'indian',
        'pacific islander', ('hispanic', 'latin'), 'white', 'other'
    )

    job = util.IndexedREMap(
        'student', ('art', 'music', 'writing'), ('banking', 'finance'),
        'administration', 'technology', 'construction', 'education',
        ('entertainment', 'media'), 'management', 'hospitality', 'law',
        'medicine', 'military', ('politics', 'government'),
        ('sales', 'marketing'), ('science', 'engineering'),
        'transportation', 'unemployed', 'other', 'rather not say',
        'retired'
    )

    status = util.IndexedREMap(
        'single', 'seeing someone', 'married', 'in an open relationship'
    )

    monogamy = util.IndexedREMap('(:?[^\-]monogamous)|(:?^monogamous)',
                                   'non-monogamous')
    strictness = util.IndexedREMap('mostly', 'strictly')

    # Doesn't want kids is index 6 for some reason.
    has_kids = util.IndexedREMap('has a kid', 'has kids', (), (), (),
                                 "doesn't have kids")
    wants_kids = util.IndexedREMap('might want', 'wants', "doesn't want")

    education_status = util.IndexedREMap('graduated', 'working', 'dropped out',
                                         default=0)
    education_level = util.IndexedREMap(
        'high ?school', 'two[- ]year college', 'university', 'college',
        'masters program', 'law school', 'med school', 'ph.d program',
        'space camp'
    )

    religion = util.IndexedREMap('agnosticism', 'atheism', 'christianity',
                                 'judaism', 'catholicism', 'islam', 'hinduism',
                                 'buddhism', 'other', default=1, offset=2)
    seriousness = util.IndexedREMap('very serious', 'somewhat serious',
                                    'not too serious', 'laughing')

    sign = util.IndexedREMap(
        'aquarius', 'pisces', 'aries', 'taurus', 'gemini', 'cancer', 'leo',
        'virgo', 'libra', 'scorpio', 'sagittarius', 'capricorn',
    )
    importance = util.IndexedREMap(
        "doesn't matter", "matters alot", "fun to think about"
    )

    dogs = util.REMap.from_string_pairs(
        (('dislikes dogs', 3), ('has dogs', 1), ('likes dogs', 2)),
        default=0
    )
    cats = util.REMap.from_string_pairs(
        (('dislikes cats', 3), ('has cats', 1), ('likes cats', 2)),
        default=0
    )
    language_level = util.IndexedREMap('fluently', 'okay', 'poorly')

    diet_strictness = util.IndexedREMap('[Mm]ostly', '[Ss]trictly')

    diet = util.IndexedREMap('anything', 'vegetarian', 'vegan',
                                  'kosher', 'halal', 'other')
    income = util.IndexedREMap(
        'less than $20,000', '$20,000-$30,000', '$30,000-$40,000',
        '$40,000-$50,000', '$50,000-$60,000', '$60,000-$70,000',
        '$70,000-$80,000', '$80,000-$100,000', '$100,000-$150,000',
        '$150,000-$250,000', '$250,000-$500,000', '$500,000-$1,000,000',
        'More than $1,000,000'
    )


class MappingUpdater(object):

    def __init__(self, mapping):
        self.mapping = mapping

    def __call__(self, id_name, value):
        if isinstance(value, six.string_types):
            value = value.lower() # :/
        return {id_name: self.mapping[value]}


class SimpleFilterBuilder(object):

    def __init__(self, filter_number, mapping, offset=0):
        self.filter_number = filter_number
        self.mapping = mapping

    def get_number(self, values):
        acc = 0
        for value in values:
            acc += 2 ** int(self.mapping[value])
        return acc

    def get_filter(self, values):
        return '{0},{1}'.format(self.filter_number, self.get_number(values))

    __call__ = get_filter


class filters(six.with_metaclass(util.GetAttrGetItem)):

    bodytype = SimpleFilterBuilder(30, maps.bodytype)
    smokes = SimpleFilterBuilder(11, maps.smokes)
    drinks = SimpleFilterBuilder(12, maps.drinks)
    drugs = SimpleFilterBuilder(13, maps.drugs)
    education_level = SimpleFilterBuilder(19, maps.education_level)
    job = SimpleFilterBuilder(15, maps.job)
    income = SimpleFilterBuilder(14, maps.income)
    religion = SimpleFilterBuilder(8, maps.religion)
    monogamy = SimpleFilterBuilder(73, maps.monogamy)
    diet = SimpleFilterBuilder(54, maps.diet)
    sign = SimpleFilterBuilder(21, maps.sign)
    ethnicities = SimpleFilterBuilder(9, maps.ethnicities)
    dogs = SimpleFilterBuilder(16, maps.dogs)
    cats = SimpleFilterBuilder(17, maps.cats)


imperial_re = re.compile(u"([0-9])['′\u2032] ?([0-9][0-1]?)[\"″\u2033]",
                                 flags=re.UNICODE)
metric_re = re.compile(u"([0-2]\.[0-9]{1,2})m")


sep_replacements = ('\\', '/', '.', '-', ' ', '$', ',', '(', ')')


gentation_to_number = {
    'women who like men': '34',
    'women': '34',
    'men who like women': '17',
    'women who like women': '40',
    'men who like men': '20',
    'men and women who like bi men': '54',
    'men and women who like bi women': '57',
    'both who like bi men': '54',
    'both who like bi women': '57',
    'straight women only': '2',
    'straight men only': '1',
    'gay women only': '8',
    'gay men only': '4',
    'bi women only': '32',
    'bi men only': '16',
    'bi men and women': '48',
    'everybody': '63',
    'girls who like guys': '34',
    'guys who like girls': '17',
    'girls who like girls': '40',
    'guys who like guys': '20',
    'guys and girls who like bi guys': '54',
    'guys and girls who like bi girls': '57',
    'both who like bi guys': '54',
    'both who like bi girls': '57',
    'straight girls only': '2',
    'straight guys only': '1',
    'gay girls only': '8',
    'gay guys only': '4',
    'bi girls only': '32',
    'bi guys only': '16',
    'bi guys and girls': '48',
    'everybody': '63',
    '': '63'
}


looking_for_re_numbers = ((re.compile("[Ff]riends"), 1),
                          (re.compile("[Ll]ong.*[Dd]ating"), 2),
                          (re.compile("[Ss]hort.*[Dd]ating"), 3),
                          (re.compile("[Ss]ex"), 6))


def inches_to_centimeters(inches):
    return inches * 2.54


def get_height_filter(height_min=None, height_max=None):
    min_int = 0
    max_int = 99999
    if isinstance(height_min, six.string_types):
        min_int = 100 * parse_height_string(height_min)
    elif isinstance(height_min, int):
        min_int = 100 * inches_to_centimeters(height_min)
    if isinstance(height_max, six.string_types):
        max_int = 100 * parse_height_string(height_max)
    elif isinstance(height_max, int):
        max_int = 100 * inches_to_centimeters(height_max)
    return '10,{0},{1}'.format(min_int, max_int)


def parse_height_string(height_string):
    if not height_string:
        return 0
    match = imperial_re.search(height_string)
    if match:
        return int(round(inches_to_centimeters(int(match.group(1)) * 12 +
                                               int(match.group(2)))))
    match = metric_re.search(height_string)
    if match:
        meters = float(match.group(1))
        centimeters = meters * 100
        return int(round(centimeters))
    else:
        raise ValueError("The provided height did not match any of "
                         "the accepted formats.")


def get_kids_filter(has_kids=(), wants_kids=()):
    return '18,{0}'.format(get_kids_int(has_kids, wants_kids))


def get_kids_int(has_kids, wants_kids):
    has_kids = util.makelist(has_kids or ())
    wants_kids = util.makelist(wants_kids)
    wants_was_unknown = False
    has_was_unknown = False
    kids_int = 0
    has_kids_ints = [maps.has_kids[matchable] for matchable in has_kids]
    if len(has_kids_ints) == 0 or sum(has_kids_ints) == 0:
        has_kids_ints = list(maps.has_kids.values()) + [0]
        has_was_unknown = True
    wants_kids_ints = [maps.wants_kids[matchable] for matchable in wants_kids]
    if len(wants_kids_ints) == 0 or sum(wants_kids_ints) == 0:
        wants_kids_ints = list(maps.wants_kids.values()) + [0]
        wants_was_unknown = True
    wants_kids_ints = [v * 8 for v in wants_kids_ints]

    kids_int += sum(2 ** (hk_int + wk_int)
                    for hk_int, wk_int
                    in itertools.product(has_kids_ints, wants_kids_ints))

    if not (wants_was_unknown or has_was_unknown):
        # ... Why? Apparently this is needed...
        kids_int = subtract_has_kids_exponents(kids_int)

    # This is super crazy and gross. I'm not really sure why they do this.
    if not wants_was_unknown and has_was_unknown:
        if 3 in [maps.wants_kids[matchable] for matchable in wants_kids]:
            kids_int += 48

    return kids_int


def subtract_has_kids_exponents(value):
    has_kids_values = maps.has_kids.values()
    for exponent in yield_exponents_of_two(value):
        if exponent in has_kids_values:
            value -= exponent
    return value


def yield_exponents_of_two(value):
    power = 0
    while value > 0:
        if value & 0b1:
            yield power
        power += 1
        value >>= 1


def get_language_query(language):
    return '22,{0}'.format(language_map[language.lower()])


hour = 60*60
join_date_string_to_int = {
    'hour': hour,
    'day': hour*24,
    'week': hour*24*7,
    'month': hour*24*30,
    'year': 365*24*hour
}


def get_join_date_filter(join_date):
    date_int = 0
    if isinstance(join_date, str):
        if join_date in join_date_string_to_int:
            date_int = join_date_string_to_int[join_date]
        else:
            date_int = int(join_date)
    return '6,{0}'.format(date_int)


def get_question_filter(question, question_answers=None):
    question_id = question
    if isinstance(question, UserQuestion):
        question_id = question.id
    if question_answers is None:
        question_answers = [answer_option.id
                            for answer_option in question.answer_options
                            if answer_option.is_match]
    answers_int = 0
    for answer_int in question_answers:
        answers_int += 2 ** answer_int
    return '65,{0},{1}'.format(question_id, answers_int)


language_map_2 = {
    'afrikaans': 'af',
    'albanian': 'sq',
    'arabic': 'ar',
    'armenian': 'hy',
    'basque': 'eu',
    'belarusan': 'be',
    'bengali': 'bn',
    'breton': 'br',
    'bulgarian': 'bg',
    'catalan': 'ca',
    'cebuano': '11',
    'chechen': 'ce',
    'chinese': 'zh',
    'c++': '71',
    'croatian': 'hr',
    'czech': 'cs',
    'danish': 'da',
    'dutch': 'nl',
    'english': 'en',
    'esperanto': 'eo',
    'estonian': 'et',
    'farsi': '20',
    'finnish': 'fi',
    'french': 'fr',
    'frisian': '23',
    'georgian': '24',
    'german': 'de',
    'greek': 'el',
    'gujarati': 'gu',
    'ancient greek': '27',
    'hawaiian': '28',
    'hebrew': 'he',
    'hindi': 'hi',
    'hungarian': 'hu',
    'icelandic': 'is',
    'ilongo': '32',
    'indonesian': 'id',
    'irish': 'ga',
    'italian': 'it',
    'japanese': 'ja',
    'khmer': '37',
    'korean': 'ko',
    'latin': 'la',
    'latvian': 'lv',
    'lisp': '72',
    'lithuanian': 'lt',
    'malay': 'ms',
    'maori': 'mi',
    'mongolian': 'mn',
    'norwegian': 'no',
    'occitan': 'oc',
    'other': '73',
    'persian': 'fa',
    'polish': 'pl',
    'portuguese': 'pt',
    'romanian': 'ro',
    'rotuman': '51',
    'russian': 'ru',
    'sanskrit': 'sa',
    'sardinian': '54',
    'serbian': 'sr',
    'sign language': '1',
    'slovak': 'sk',
    'slovenian': 'sl',
    'spanish': 'es',
    'swahili': 'sw',
    'swedish': 'sv',
    'tagalog': 'tl',
    'tamil': 'ta',
    'thai': 'th',
    'tibetan': 'bo',
    'turkish': 'tr',
    'ukrainian': 'uk',
    'urdu': 'ur',
    'vietnamese': 'vi',
    'welsh': 'cy',
    'yiddish': 'yi'
}


language_map = {
    'afrikaans': '2',
    'albanian': '3',
    'ancient greek': '27',
    'arabic': '4',
    'armenian': '76',
    'basque': '5',
    'belarusan': '6',
    'bengali': '7',
    'breton': '8',
    'bulgarian': '9',
    'c++': '71',
    'catalan': '10',
    'cebuano': '11',
    'chechen': '12',
    'chinese': '13',
    'croatian': '14',
    'czech': '15',
    'danish': '16',
    'dutch': '17',
    'english': '74',
    'esperanto': '18',
    'estonian': '19',
    'farsi': '20',
    'finnish': '21',
    'french': '22',
    'frisian': '23',
    'georgian': '24',
    'german': '25',
    'greek': '26',
    'gujarati': '77',
    'hawaiian': '28',
    'hebrew': '29',
    'hindi': '75',
    'hungarian': '30',
    'icelandic': '31',
    'ilongo': '32',
    'indonesian': '33',
    'irish': '34',
    'italian': '35',
    'japanese': '36',
    'khmer': '37',
    'korean': '38',
    'lisp': '72',
    'latin': '39',
    'latvian': '40',
    'lithuanian': '41',
    'malay': '42',
    'maori': '43',
    'mongolian': '44',
    'norwegian': '45',
    'occitan': '46',
    'other': '73',
    'persian': '47',
    'polish': '48',
    'portuguese': '49',
    'romanian': '50',
    'rotuman': '51',
    'russian': '52',
    'sanskrit': '53',
    'sardinian': '54',
    'serbian': '55',
    'sign language': '1',
    'slovak': '56',
    'slovenian': '57',
    'spanish': '58',
    'swahili': '59',
    'swedish': '60',
    'tagalog': '61',
    'tamil': '62',
    'thai': '63',
    'tibetan': '64',
    'turkish': '65',
    'ukrainian': '66',
    'urdu': '67',
    'vietnamese': '68',
    'welsh': '69',
    'yiddish': '70',
}



import pkg_resources

from invoke import Program

from .attractiveness_finder import AttractivenessFinder
from .photo import PhotoUploader
from .session import Session
from .statistics import Statistics
from .user import User
from . import tasks
from .util import save_file


__version__ = pkg_resources.get_distribution('okcupyd').version

@property
def always_add_task_args_initial_context(self):
    from invoke.parser import ParserContext
    args = self.core_args()
    args += self.task_args()
    return ParserContext(args=args)


def interactive():
    Program.initial_context = always_add_task_args_initial_context
    return Program(name="OKCupyd", version=__version__, namespace=tasks.ns).run()


__all__ = ('User', 'AttractivenessFinder', 'Statistics',
           'save_file', 'interactive', 'PhotoUploader', 'Session',
           '__version__')



from invoke import task

from okcupyd.util import curry


def build_task_factory(ns):
    @curry
    def task_decorator(function, **kwargs):
        name = kwargs.pop("name", None)
        default = kwargs.pop("default", None)
        created_task = task(function, **kwargs)
        ns.add_task(created_task, name=name, default=default)
        return created_task
    return task_decorator



import copy
import inspect
import logging
import os
import zlib

from six.moves import urllib
import simplejson
import vcr
import wrapt

from okcupyd import settings
from okcupyd import util


log = logging.getLogger(__name__)

TESTING_USERNAME = 'username'
TESTING_PASSWORD = 'password'
WBITS = 16 + zlib.MAX_WBITS


SHOULD_SCRUB = False
REPLACEMENTS = []
REMOVE_OLD_CASSETTES = False


@wrapt.decorator
def check_should_scrub(function, instance, args, kwargs):
    if SHOULD_SCRUB:
        return function(*args)
    else:
        return args[0] # The request or response


@util.curry
def remove_headers(request, headers_to_remove=()):
    headers = copy.copy(request.headers)
    headers_to_remove = [h.lower() for h in headers_to_remove]
    keys = [k for k in headers if k.lower() in headers_to_remove]
    if keys:
        for k in keys:
            headers.pop(k)
        request.headers = headers
    return request


def scrub_request_body(request):
    if urllib.parse.urlsplit(request.uri).path == '/login':
        request.body = scrub_query_string(request.body)
    request.uri = scrub_uri(request.uri)
    return request


def scrub_uri(uri):
    replaced = util.replace_all_case_insensitive(uri, settings.USERNAME,
                                                 TESTING_USERNAME)
    return util.replace_all_case_insensitive(replaced, settings.PASSWORD,
                                             TESTING_PASSWORD)


def scrub_query_string(query_string):
    request_dict = urllib.parse.parse_qs(query_string)
    if 'password' not in request_dict:
        return query_string

    for key in request_dict:
        request_dict[key] = request_dict[key][0]
    request_dict['username'] = TESTING_USERNAME
    request_dict['password'] = TESTING_PASSWORD
    return urllib.parse.urlencode(request_dict)


def gzip_string(incoming):
    if isinstance(incoming, str) and bytes is not str:
        incoming = bytes(incoming, 'utf8')
    else:
        incoming = incoming.encode('utf8')
    compress_object = zlib.compressobj(6, zlib.DEFLATED, WBITS)
    start = compress_object.compress(incoming)
    end = compress_object.flush()
    if not isinstance(start, str):
        return start + end
    return ''.join([start, end])


def scrub_response_headers(response):
    for item in ('location', 'Location'):
        if item in response['headers']:
            response['headers'][item] = [scrub_uri(uri)
                                         for uri in response['headers'][item]]
    return response


def replace_json_fields(body):
    try:
        response_dict = simplejson.loads(body)
    except:
        return body

    if 'screenname' not in response_dict:
        return body
    if response_dict['screenname'] is not None:
        response_dict['screenname'] = TESTING_USERNAME
    response_dict['userid'] = 1
    response_dict['thumbnail'] = ''
    return simplejson.dumps(response_dict)


def scrub_response(response):
    if not SHOULD_SCRUB:
        return response
    response = response.copy()
    response = scrub_response_headers(response)

    body = response['body']['string']
    try:
        body = zlib.decompress(response['body']['string'], WBITS).decode('utf8')
    except:
        should_recompress = False
    else:
        should_recompress = True

    body = replace_json_fields(body)
    body = util.replace_all_case_insensitive(body, settings.USERNAME,
                                             TESTING_USERNAME)

    if should_recompress:
        body = gzip_string(body)
    response['body']['string'] = body
    return response


before_record = check_should_scrub(util.compose(
    scrub_request_body, remove_headers(headers_to_remove=(
        'Set-Cookie',
        'Cookie'
    ))
))


def _maybe_decode(maybe_bytes):
    try:
        return maybe_bytes.decode('utf-8')
    except (AttributeError, UnicodeDecodeError):
        return maybe_bytes


def _match_search_query(left, right):
    left_filter = set([value for param_name, value in left
                       if 'filter' in _maybe_decode(param_name)])
    right_filter = set([value for param_name, value in right
                        if 'filter' in _maybe_decode(param_name)])
    left_rest = set([(param_name, value) for param_name, value in left
                     if 'filter' not in _maybe_decode(param_name)])
    right_rest = set([(param_name, value) for param_name, value in right
                      if 'filter' not in _maybe_decode(param_name)])

    try:
        log.info(simplejson.dumps(
            {
                'filter_differences': list(
                    left_filter.symmetric_difference(right_filter)
                ),
                'rest_differences': list(
                    left_rest.symmetric_difference(right_rest)
                ),
            }, encoding='utf-8'
        ))
    except Exception as e:
        log.warning(e)

    return left_filter == right_filter and left_rest == right_rest


def match_search_query(left, right):
    return _match_search_query(left.query, right.query)


def body_as_query_string(left, right):
    if left.path == right.path and 'ajaxuploader' in left.path:
        return True # We can't seem to handle matching photo uploads likely
        # because of requests internals.
    try:
        left_qs_items = list(urllib.parse.parse_qs(left.body).items())
        right_qs_items = list(urllib.parse.parse_qs(right.body).items())
    except Exception as exc:
        log.debug(exc)
        return left.body == right.body
    else:
        left_qs_items = [(k, tuple(v)) for k, v in left_qs_items]
        right_qs_items = [(k, tuple(v)) for k, v in right_qs_items]
        return _match_search_query(left_qs_items, right_qs_items)


cassette_library_directory = os.path.join(os.path.dirname(os.path.dirname(__file__)),
                                          'tests', 'vcr_cassettes')
okcupyd_vcr = vcr.VCR(match_on=('path', 'method', 'match_search_query',
                                'body_as_query_string'),
                      before_record=(before_record,),
                      before_record_response=scrub_response,
                      cassette_library_dir=cassette_library_directory,
                      path_transformer=vcr.VCR.ensure_suffix('.yaml'))
okcupyd_vcr.register_matcher('body_as_query_string', body_as_query_string)
okcupyd_vcr.register_matcher('match_search_query', match_search_query)
match_on_no_body = list(filter(lambda x: 'body' not in x, okcupyd_vcr.match_on))


@wrapt.adapter_factory
def add_request_to_signature(function):
    argspec = inspect.getargspec(function)
    return inspect.ArgSpec(argspec.args + ['request'], argspec.varargs, argspec.keywords, argspec.defaults)


@wrapt.decorator(adapter=add_request_to_signature)
def skip_if_live(function, instance, args, kwargs):
    request = kwargs.pop('request')
    if request.config.getoption('skip_vcrpy'):
        log.debug("Skipping {0} because vcrpy is being skipped.".format(
            function.__name__
        ))
    else:
        return function(*args, **kwargs)


use_cassette = okcupyd_vcr.use_cassette
